import basix


def to_x(p):
    if len(p) == 1:
        return 100 * p[0]
    if len(p) == 2:
        return 100 * p[0]
    if len(p) == 3:
        return 100 * p[0] + 30 * p[1]


def to_y(p):
    if len(p) == 1:
        return 120
    if len(p) == 2:
        return 120 - 100 * p[1]
    if len(p) == 3:
        return 120 - 100 * p[2] - 40 * p[1]


for shape in ["interval", "triangle", "tetrahedron",
              "quadrilateral", "hexahedron", "prism", "pyramid"]:
    cell_type = getattr(basix.CellType, shape)
    geometry = basix.geometry(cell_type)
    topology = basix.topology(cell_type)

    yadd = 0
    width = 100
    if shape == "interval":
        yadd = -100
    if shape == "hexahedron":
        yadd = 40
        width = 140
    if shape == "pyramid":
        width = 140
    if shape == "prism":
        yadd = 40

    svg = ""

    if geometry.shape[1] == 1:
        svg += (f"<line x1='20' y1='{120 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<line x1='40' y1='{115 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<line x1='40' y1='{125 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<text x='60' y='{120 + yadd}' fill='#000000' dy='.3em' "
                "style='font-family:\"Libertinus Serif Semibold Italic\";font-size:20px'>x</text>\n")
    elif geometry.shape[1] == 2:
        svg += (f"<line x1='20' y1='{120 + yadd}' x2='20' y2='{90 + yadd}' />\n"
                f"<line x1='15' y1='{100 + yadd}' x2='20' y2='{90 + yadd}' />\n"
                f"<line x1='25' y1='{100 + yadd}' x2='20' y2='{90 + yadd}' />\n"
                f"<line x1='20' y1='{120 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<line x1='40' y1='{115 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<line x1='40' y1='{125 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<text x='60' y='{120 + yadd}' fill='#000000' dy='.3em' "
                "style='font-family:\"Libertinus Serif Semibold Italic\";font-size:20px'>x</text>\n"
                f"<text x='20' y='{75 + yadd}' fill='#000000' dy='.3em' "
                "style='font-family:\"Libertinus Serif Semibold Italic\";font-size:20px'>y</text>\n")
    elif geometry.shape[1] == 3:
        svg += (f"<line x1='20' y1='{120 + yadd}' x2='20' y2='{90 + yadd}' />\n"
                f"<line x1='15' y1='{100 + yadd}' x2='20' y2='{90 + yadd}' />\n"
                f"<line x1='25' y1='{100 + yadd}' x2='20' y2='{90 + yadd}' />\n"
                f"<line x1='20' y1='{120 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<line x1='40' y1='{115 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<line x1='40' y1='{125 + yadd}' x2='50' y2='{120 + yadd}' />\n"
                f"<line x1='20' y1='{120 + yadd}' x2='44' y2='{102 + yadd}' />\n"
                f"<line x1='33' y1='{104 + yadd}' x2='44' y2='{102 + yadd}' />\n"
                f"<line x1='39' y1='{112 + yadd}' x2='44' y2='{102 + yadd}' />\n"
                f"<text x='60' y='{120 + yadd}' fill='#000000' dy='.3em' "
                "style='font-family:\"Libertinus Serif Semibold Italic\";font-size:20px'>x</text>\n"
                f"<text x='52' y='{91 + yadd}' fill='#000000' dy='.3em' "
                "style='font-family:\"Libertinus Serif Semibold Italic\";font-size:20px'>y</text>\n"
                f"<text x='20' y='{75 + yadd}' fill='#000000' dy='.3em' "
                "style='font-family:\"Libertinus Serif Semibold Italic\";font-size:20px'>z</text>\n")

    xpos = 100
    for dim, es in enumerate(topology):
        lines = []
        for e in topology[1]:
            lines.append(f"<line x1='{xpos + to_x(geometry[e[0]])}' y1='{yadd + to_y(geometry[e[0]])}'"
                         f" x2='{xpos + to_x(geometry[e[1]])}' y2='{yadd + to_y(geometry[e[1]])}' />\n")

        entities = []
        for n_e, e in enumerate(es):
            points = [geometry[i] for i in e]
            mid = sum(points) / len(points)
            e_svg = f"<circle cx='{xpos + to_x(mid)}' cy='{yadd + to_y(mid)}' r='12px' />"
            e_svg += f"<text x='{xpos + to_x(mid)}' y='{yadd + to_y(mid)}' dy='.3em'"
            if n_e >= 10:
                e_svg += " style='font-size:10px'"
            e_svg += f">{n_e}</text>\n"
            entities.append(e_svg)

        if shape == "hexahedron" and dim == 1:
            svg += "".join([lines[i] for i in [5, 6, 7, 11]])
            svg += "".join([entities[i] for i in [5, 6, 7, 11]])
            svg += "".join([lines[i] for i in [1, 3, 9, 10]])
            svg += "".join([entities[i] for i in [1, 3, 9, 10]])
            svg += "".join([lines[i] for i in [0, 2, 4, 8]])
            svg += "".join([entities[i] for i in [0, 2, 4, 8]])
        elif shape == "hexahedron" and dim == 2:
            svg += "".join([lines[i] for i in [5, 6, 7, 11]])
            svg += "".join([entities[i] for i in [4]])
            svg += "".join([lines[i] for i in [1, 3, 9, 10]])
            svg += "".join([entities[i] for i in [0, 2, 3, 5]])
            svg += "".join([lines[i] for i in [0, 2, 4, 8]])
            svg += "".join([entities[i] for i in [1]])
        elif shape == "tetrahedron" and dim == 1:
            svg += "".join([lines[i] for i in [0, 2, 4]])
            svg += "".join([entities[i] for i in [0, 2, 4]])
            svg += "".join([lines[i] for i in [1, 3, 5]])
            svg += "".join([entities[i] for i in [1, 3, 5]])
        elif shape == "tetrahedron" and dim == 2:
            svg += "".join([lines[i] for i in [0, 2, 4]])
            svg += "".join([entities[i] for i in [0, 1, 3]])
            svg += "".join([lines[i] for i in [1, 3, 5]])
            svg += "".join([entities[i] for i in [2]])
        elif shape == "prism" and dim == 1:
            svg += lines[5]
            svg += entities[5]
            svg += "".join([lines[i] for i in [1, 2, 7, 8]])
            svg += "".join([entities[i] for i in [1, 2, 7, 8]])
            svg += "".join([lines[i] for i in [0, 3, 4, 6]])
            svg += "".join([entities[i] for i in [0, 3, 4, 6]])
        elif shape == "pyramid" and dim == 1:
            svg += lines[2]
            svg += entities[2]
            svg += "".join([lines[i] for i in [1, 3, 6, 7]])
            svg += "".join([entities[i] for i in [1, 3, 6, 7]])
            svg += "".join([lines[i] for i in [0, 4, 5]])
            svg += "".join([entities[i] for i in [0, 4, 5]])
        elif shape == "pyramid" and dim == 2:
            svg += lines[5]
            svg += entities[3]
            svg += "".join([lines[i] for i in [1, 3, 6, 7]])
            svg += "".join([entities[i] for i in [0, 2, 4]])
            svg += "".join([lines[i] for i in [0, 4, 2]])
            svg += "".join([entities[i] for i in [1]])
        elif shape == "pyramid" and dim == 3:
            svg += "".join([lines[i] for i in [1, 3, 5, 6, 7]])
            svg += entities[0]
            svg += "".join([lines[i] for i in [0, 2, 4]])
        else:
            svg += "\n".join(lines)
            svg += "\n".join(entities)

        xpos += width + 70

    with open(f"{shape}_numbering.svg", "w") as f:
        f.write(f"<svg width='{xpos - 50}' height='{140 + yadd}'>\n")
        f.write("<style type=\"text/css\"><![CDATA[\n"
                "  text { text-anchor:middle; font-size:15px; font-family: \"Lato Bold\" }\n"
                "  line { stroke-width: 4px; stroke-linecap: round; stroke: #000000 }\n"
                "  circle { fill: #FFFFFF; stroke: #000000; stroke-width:4px}\n"
                "  rect { fill: #FFFFFF; stroke: #FFFFFF; stroke-width:4px}\n"
                "]]></style>\n")
        f.write(f"<rect x='0' y='0' width='{xpos - 50}' height='{140 + yadd}' />\n")

        f.write(svg)
        f.write("</svg>")



# Copyright (c) 2022 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import numpy as np


def test_element_equality():
    p1 = basix.create_element(basix.ElementFamily.P, basix.CellType.triangle, 1)
    p1_again = basix.create_element(basix.ElementFamily.P, basix.CellType.triangle, 1)
    rt1 = basix.create_element(basix.ElementFamily.RT, basix.CellType.triangle, 1)
    p1_quad = basix.create_element(basix.ElementFamily.P, basix.CellType.quadrilateral, 1)
    p4_gll = basix.create_element(basix.ElementFamily.P, basix.CellType.triangle, 4, basix.LagrangeVariant.gll_warped)
    p4_equi = basix.create_element(basix.ElementFamily.P, basix.CellType.triangle, 4, basix.LagrangeVariant.equispaced)

    assert p1 == p1
    assert p1 == p1_again
    assert p1 != p4_gll
    assert p1 != p4_equi
    assert p4_gll != p4_equi
    assert p1 != p1_quad
    assert p1 != rt1


def test_custom_element_equality():
    p1 = basix.create_element(basix.ElementFamily.P, basix.CellType.triangle, 1)
    cr = basix.create_element(basix.ElementFamily.CR, basix.CellType.triangle, 1)

    wcoeffs = np.eye(3)
    z = np.zeros((0, 2))
    x = [[np.array([[0., 0.]]), np.array([[1., 0.]]), np.array([[0., 1.]])], [z, z, z], [z], []]
    z = np.zeros((0, 1, 0))
    M = [[np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])], [z, z, z], [z], []]

    p1_custom = basix.create_custom_element(
        basix.CellType.triangle, [], wcoeffs,
        x, M, basix.MapType.identity, False, 1, 1)
    p1_custom_again = basix.create_custom_element(
        basix.CellType.triangle, [], wcoeffs,
        x, M, basix.MapType.identity, False, 1, 1)

    wcoeffs = np.eye(3)
    z = np.zeros((0, 2))
    x = [[z, z, z], [np.array([[.5, .5]]), np.array([[0., .5]]), np.array([[.5, 0.]])], [z], []]
    z = np.zeros((0, 1, 0))
    M = [[z, z, z], [np.array([[[1.]]]), np.array([[[1.]]]), np.array([[[1.]]])], [z], []]

    cr_custom = basix.create_custom_element(
        basix.CellType.triangle, [], wcoeffs,
        x, M, basix.MapType.identity, False, 1, 1)

    assert p1_custom == p1_custom_again
    assert p1_custom != cr_custom
    assert p1_custom != cr
    assert p1_custom == p1
    assert cr_custom == cr



"""Helper functions for writing DOLFINx custom kernels using Numba."""

import numpy as _np
import numpy.typing as _npt
from typing import List as _ListT
from typing import Dict as _Dict
try:
    import numba as _numba
    from numba.typed import List as _List
    from numba.core import types as _types
except ImportError:
    raise RuntimeError("You must have Numba installed to use the Numba helper functions.")


@_numba.njit
def apply_dof_transformation(
    tdim: int, edge_count: int, face_count: int, entity_transformations: _ListT[int], entity_dofs: _ListT[int],
    data: _np.array, cell_info: int, face_types: _ListT[str]
):
    """Apply dof transformations to some data.

    Args:
        tdim: The topological dimension of the cell.
        edge_count: The number of edges the cell has.
        face_count: The number of faces the cell has.
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
        face_types: A list of strings giving the shapes of the faces of the cell.
    """
    if tdim >= 2:
        if tdim == 3:
            face_start = 3 * face_count
        else:
            face_start = 0

        dofstart = 0
        for i in entity_dofs[0]:
            dofstart += i
        # NOTE: Copy array to make numba compilation faster (contiguous array assumption)
        edge_reflection = entity_transformations["interval"][0].copy()
        for e in range(edge_count):
            edofs = entity_dofs[1][e]
            if edofs == 0:
                continue
            if cell_info >> (face_start + e) & 1:
                data[dofstart:dofstart+edofs] = _np.dot(edge_reflection, data[dofstart:dofstart+edofs])
            dofstart += edofs

        if tdim == 3:
            for f in range(face_count):
                face_rotation = entity_transformations[face_types[f]][0].copy()
                face_reflection = entity_transformations[face_types[f]][1].copy()
                fdofs = entity_dofs[2][f]
                if fdofs == 0:
                    continue
                if cell_info >> (3 * f) & 1:
                    data[dofstart:dofstart+fdofs] = _np.dot(face_reflection, data[dofstart:dofstart+fdofs])
                for _ in range(cell_info >> (3 * f + 1) & 3):
                    data[dofstart:dofstart+fdofs] = _np.dot(face_rotation, data[dofstart:dofstart+fdofs])
                dofstart += fdofs


@_numba.njit
def apply_dof_transformation_interval(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some data on an interval.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    return


@_numba.njit
def apply_dof_transformation_triangle(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some data on a triangle.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation(2, 3, 1, entity_transformations, entity_dofs,
                             data, cell_info, _List.empty_list(_types.string))


@_numba.njit
def apply_dof_transformation_quadrilateral(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some data on an quadrilateral.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation(2, 4, 1, entity_transformations, entity_dofs,
                             data, cell_info, _List.empty_list(_types.string))


@_numba.njit
def apply_dof_transformation_tetrahedron(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some data on a tetrahedron.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation(3, 6, 4, entity_transformations, entity_dofs,
                             data, cell_info, _List(["triangle"] * 4))


@_numba.njit
def apply_dof_transformation_hexahedron(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some data on a hexahedron.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation(3, 12, 6, entity_transformations, entity_dofs,
                             data, cell_info, _List(["quadrilateral"] * 6))


@_numba.njit
def apply_dof_transformation_prism(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some data on an prism.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation(3, 9, 5, entity_transformations, entity_dofs,
                             data, cell_info, _List(["triangle"] + ["quadrilateral"] * 4 + ["triangle"]))


@_numba.njit
def apply_dof_transformation_pyramid(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some data on an prism.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation(3, 8, 5, entity_transformations, entity_dofs,
                             data, cell_info, _List(["quadrilateral"] + ["triangle"] * 4))


@_numba.njit
def apply_dof_transformation_to_transpose(
    tdim: int, edge_count: int, face_count: int, entity_transformations: _ListT[int], entity_dofs: _ListT[int],
    data: _np.array, cell_info: int, face_types: _ListT[str]
):
    """Apply dof transformations to some transposed data.

    Args:
        tdim: The topological dimension of the cell.
        edge_count: The number of edges the cell has.
        face_count: The number of faces the cell has.
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
        face_types: A list of strings giving the shapes of the faces of the cell.
    """
    transposed_data = data.transpose().copy()
    apply_dof_transformation(tdim, edge_count, face_count, entity_transformations, entity_dofs,
                             transposed_data, cell_info, face_types)
    data[:] = transposed_data.transpose()


@_numba.njit
def apply_dof_transformation_to_transpose_interval(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some transposed data on an interval.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    return


@_numba.njit
def apply_dof_transformation_to_transpose_triangle(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some transposed data on a triangle.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation_to_transpose(2, 3, 1, entity_transformations, entity_dofs,
                                          data, cell_info, _List.empty_list(_types.string))


@_numba.njit
def apply_dof_transformation_to_transpose_quadrilateral(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some transposed data on an quadrilateral.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation_to_transpose(2, 4, 1, entity_transformations, entity_dofs,
                                          data, cell_info, _List.empty_list(_types.string))


@_numba.njit
def apply_dof_transformation_to_transpose_tetrahedron(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some transposed data on a tetrahedron.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation_to_transpose(3, 6, 4, entity_transformations, entity_dofs,
                                          data, cell_info, _List(["triangle"] * 4))


@_numba.njit
def apply_dof_transformation_to_transpose_hexahedron(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some transposed data on a hexahedron.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation_to_transpose(3, 12, 6, entity_transformations, entity_dofs,
                                          data, cell_info, _List(["quadrilateral"] * 6))


@_numba.njit
def apply_dof_transformation_to_transpose_prism(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some transposed data on an prism.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation_to_transpose(3, 9, 5, entity_transformations, entity_dofs,
                                          data, cell_info, _List(["triangle"] + ["quadrilateral"] * 4 + ["triangle"]))


@_numba.njit
def apply_dof_transformation_to_transpose_pyramid(
    entity_transformations: _Dict[str, _npt.NDArray[_np.float64]], entity_dofs: _Dict[str, _npt.NDArray[_np.int32]],
    data: _np.array, cell_info: int
):
    """Apply dof transformations to some transposed data on an prism.

    Args:
        entity_transformations: The DOF transformations for each entity.
        entity_dofs: The number of DOFs on each entity.
        data: The data. This will be changed by this function.
        cell_info: An integer representing the orientations of the subentities of the cell.
    """
    apply_dof_transformation_to_transpose(3, 8, 5, entity_transformations, entity_dofs,
                                          data, cell_info, _List(["quadrilateral"] + ["triangle"] * 4))



# ====================================
# Creating and using a quadrature rule
# ====================================
#
# This demo shows how quadrature rules can be obtained from Basix and how these
# can be used to compute the integrals of functions. A quadrature rule uses a
# set of points (:math:`p_0` to :math:`p_{n-1}`) and weights (:math:`w_0` to
# :math:`w_{n-1}`), and approximates an integral as the weighted sum of the
# values of the function at these points, ie
#
# .. math::
#
#    \int f\,\mathrm{d}x \approx \sum_i w_if(p_i).
#
# First, we import Basix and Numpy.

import basix
import numpy as np
from basix import ElementFamily, CellType, LagrangeVariant

# To get a quadrature rule on a triangle, we use the function `make_quadrature`.
# This function takes two or three three inputs. We want to use the default
# quadrature, pass in two inputs: a cell type and an order. The order of the rule
# is equal to the degree of the highest degree polynomial that will be exactly
# integrated by this rule. In this example, we use an order 4 rule, so all quartic
# polynomials will be integrated exactly.
#
# `make_quadrature` returns two values: the points and the weights of the
# quadrature rule.

points, weights = basix.make_quadrature(CellType.triangle, 4)

# If we want to control the type of quadrature used, we can pass in three
# inputs to `make_quadrautre`. For example, the following code would force basix
# to use a Gauss-Jacobi quadrature rule:

points, weights = basix.make_quadrature(
    basix.QuadratureType.gauss_jacobi, CellType.triangle, 4)

# We now use this quadrature rule to integrate the functions :math:`f(x,y)=x^3y`
# and :math:`g(x,y)=x^3y^2` over the triangle. The exact values of these integrals
# are 1/120 (0.00833333333333...) and 1/420 (0.00238095238095...) respectively.
#
# As :math:`f` is a degree 4 polynomial, we expect our quadrature rule to be able
# to compute its integral exactly (within machine precision). :math:`g` on the
# other hand is a degree 5 polynomial, so its integral will not be computed exactly.
#
# We define Python functions that compute :math:`f` and :math:`g` for every point.
# These functions use features of Numpy to compute all the values at once.


def f(points):
    return points[:, 0] ** 3 * points[:, 1]


def g(points):
    return points[:, 0] ** 3 * points[:, 1] ** 2


# We can now use Numpy features to compute the integrals.

print(np.sum(weights * f(points)))
print(np.sum(weights * g(points)))

# We obtain the values 0.00833333333333334 and 0.002393509368731209. As expected,
# the integral of :math:`f` has been computed to within machine precision, while
# the integral of :math:`g` is correct to 4 decimal places, but is not exact.
#
# Integrating a basis function
# ============================
#
# We next use the quadrature rule to compute the integral of a basis function in
# a degree 3 Lagrange space. We first create the space and tabulate its basis
# functions at the quadrature points.

lagrange = basix.create_element(
    ElementFamily.P, CellType.triangle, 3, LagrangeVariant.equispaced)

values = lagrange.tabulate(0, points)

# We compute the integral of the third (note that the counting starts at 0) basis
# function in this space. We can obtain the values of this basis function from
# `values` by using the indices `[0, :, 3, 0]`. The integral can therefore
# computed as follows. As this basis function will be degree three, the result
# will again be exact (withing machine precision).

print(np.sum(weights * values[0, :, 3, 0]))



# Copyright (c) 2022 Chris Richardson, Garth Wells and Jack S. Hale
# FEniCS Project
# SPDX-License-Identifier: MIT

import pkg_resources
import pytest
import re
import basix


def is_canonical(version):
    """Check version number is canonical according to PEP0440

    From https://peps.python.org/pep-0440/#appendix-b-parsing-version-strings-with-regular-expressions"""
    return re.match(r'^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$', version) is not None  # noqa: E501


def test_version(python_version=pkg_resources.get_distribution("fenics-basix").version, cpp_version=basix.__version__):
    assert is_canonical(python_version)

    # Strip Python-specific versioning (dev, post) and compare with C++
    # versioning
    stripped_version = re.sub(r"(\.post(0|[1-9][0-9]*))", "", python_version)
    stripped_version = stripped_version.replace("dev", "")
    if stripped_version != cpp_version:
        raise RuntimeError(
            f"The version numbers of the Python ({pkg_resources.get_distribution('fenics-basix').version} "
            + f"-> {stripped_version}) and pybind11/C++ ({basix.__version__}) libraries does not match")


def test_test_version_logic():
    with pytest.raises(RuntimeError):
        test_version("0.4.2.dev0", "0.4.2")

    test_version("0.4.2.post0", "0.4.2")
    test_version("0.4.2.dev0", "0.4.2.0")



# Copyright (c) 2021 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
from .utils import parametrize_over_elements


@parametrize_over_elements(4)
def test_number_of_dofs(cell_type, degree, element_type, element_args):
    element = basix.create_element(element_type, cell_type, degree, *element_args)

    entity_dofs = element.entity_dofs
    num_entity_dofs = element.num_entity_dofs

    for entity_dofs, num_entity_dofs in zip(element.entity_dofs, element.num_entity_dofs):
        for dofs, ndofs in zip(entity_dofs, num_entity_dofs):
            assert len(dofs) == ndofs

    assert element.dim == sum(sum(i) for i in element.num_entity_dofs)


@parametrize_over_elements(4)
def test_ordering_of_dofs(cell_type, degree, element_type, element_args):
    """
    Check that DOFs are numbered in ascending order by entity.

    This assumption is required for DOF transformations to work correctly.
    """
    element = basix.create_element(element_type, cell_type, degree, *element_args)

    entity_dofs = element.entity_dofs
    dof = 0
    for entity_dofs in element.entity_dofs:
        for dofs in entity_dofs:
            for d in sorted(dofs):
                assert d == dof
                dof += 1


@parametrize_over_elements(4)
def test_closure_dofs(cell_type, degree, element_type, element_args):
    element = basix.create_element(element_type, cell_type, degree, *element_args)

    entity_dofs = element.entity_dofs
    entity_closure_dofs = element.entity_closure_dofs

    topology = basix.topology(cell_type)
    connectivity = basix.cell.sub_entity_connectivity(cell_type)

    for dim, t in enumerate(topology):
        for e, _ in enumerate(t):
            for dim2 in range(dim + 1):
                for e2 in connectivity[dim][e][dim2]:
                    for dof in entity_dofs[dim2][e2]:
                        assert dof in entity_closure_dofs[dim][e]



# Copyright (c) 2020 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import pytest
from basix import ElementFamily, CellType, LagrangeVariant, DPCVariant


def parametrize_over_elements(degree, reference=None, discontinuous=False):
    elementlist = []

    for k in range(1, degree + 1):
        # Elements on all cells
        for c in [CellType.interval, CellType.triangle, CellType.tetrahedron, CellType.quadrilateral,
                  CellType.hexahedron, CellType.prism, CellType.pyramid]:
            if k < 4:
                elementlist.append((c, ElementFamily.P, k, [LagrangeVariant.equispaced]))

        # Elements on all cells except pyramid
        for c in [CellType.interval, CellType.triangle, CellType.tetrahedron, CellType.quadrilateral,
                  CellType.hexahedron, CellType.prism]:
            elementlist.append((c, ElementFamily.P, k, [LagrangeVariant.gll_isaac]))
            elementlist.append((c, ElementFamily.P, k, [LagrangeVariant.gll_warped]))

        # Elements on all cells except prism and pyramid
        for c in [CellType.interval, CellType.triangle, CellType.tetrahedron, CellType.quadrilateral,
                  CellType.hexahedron]:
            if discontinuous:
                elementlist.append((c, ElementFamily.P, k, [LagrangeVariant.legendre]))

        # Elements on all cells except prism, pyramid and interval
        for c in [CellType.triangle, CellType.tetrahedron, CellType.quadrilateral, CellType.hexahedron]:
            elementlist.append((c, ElementFamily.N1E, k, [LagrangeVariant.legendre]))
            elementlist.append((c, ElementFamily.N2E, k, [LagrangeVariant.legendre, DPCVariant.legendre]))
            elementlist.append((c, ElementFamily.RT, k, [LagrangeVariant.legendre]))
            elementlist.append((c, ElementFamily.BDM, k, [LagrangeVariant.legendre, DPCVariant.legendre]))

        # Elements on simplex cells
        for c in [CellType.triangle, CellType.tetrahedron]:
            if k == 1:
                elementlist.append((c, ElementFamily.CR, k, []))
            elementlist.append((c, ElementFamily.Regge, k, []))

        # Elements on tensor product cells
        for c in [CellType.interval, CellType.quadrilateral, CellType.hexahedron]:
            for lv in [LagrangeVariant.equispaced, LagrangeVariant.gll_warped]:
                for dv in [DPCVariant.simplex_equispaced, DPCVariant.diagonal_gll]:
                    elementlist.append((c, ElementFamily.serendipity, k, [lv, dv]))

        # Elements on quads and hexes
        for c in [CellType.quadrilateral, CellType.hexahedron]:
            if discontinuous:
                for v in [DPCVariant.simplex_equispaced, DPCVariant.simplex_gll,
                          DPCVariant.horizontal_equispaced, DPCVariant.horizontal_gll,
                          DPCVariant.diagonal_equispaced, DPCVariant.diagonal_gll]:
                    elementlist.append((c, ElementFamily.dpc, k, [v]))

        # Bubble elements
        if k >= 2:
            elementlist.append((CellType.interval, ElementFamily.bubble, k, []))
            elementlist.append((CellType.quadrilateral, ElementFamily.bubble, k, []))
            elementlist.append((CellType.hexahedron, ElementFamily.bubble, k, []))
        if k >= 3:
            elementlist.append((CellType.triangle, ElementFamily.bubble, k, []))
        if k >= 4:
            elementlist.append((CellType.tetrahedron, ElementFamily.bubble, k, []))

    if reference is None:
        if len(elementlist) == 0:
            raise ValueError(f"No elements will be tested with reference: {reference}")
        return pytest.mark.parametrize("cell_type, element_type, degree, element_args", elementlist)
    else:
        elementlist = [(b, c, d) for a, b, c, d in elementlist if a == reference]
        if len(elementlist) == 0:
            raise ValueError(f"No elements will be tested with reference: {reference}")
        return pytest.mark.parametrize("element_type, degree, element_args", elementlist)



"""Functions to directly wrap Basix elements in UFL."""

import ufl as _ufl
from ufl.finiteelement.finiteelementbase import FiniteElementBase as _FiniteElementBase
import hashlib as _hashlib
import basix as _basix
import typing as _typing


class BasixElement(_FiniteElementBase):
    """A wrapper allowing Basix elements to be used with UFL."""

    def __init__(self, element: _basix.finite_element.FiniteElement):
        """Create a basix element."""
        super().__init__(
            element.family.name, element.cell_type.name, element.degree, None, tuple(element.value_shape),
            tuple(element.value_shape))

        if element.family == _basix.ElementFamily.custom:
            self._repr = f"custom Basix element ({_compute_signature(element)})"
        else:
            self._repr = (f"Basix element ({element.family.name}, {element.cell_type.name}, {element.degree}, "
                          f"{element.lagrange_variant.name}, {element.dpc_variant.name}, {element.discontinuous})")
        self.basix_element = element

    def mapping(self) -> str:
        """Return the map type."""
        return _map_type_to_string(self.basix_element.map_type)

    def __eq__(self, other):
        """Check if two elements are equal."""
        if isinstance(other, BasixElement):
            return self.basix_element == other.basix_element
        return False

    def __hash__(self):
        """Return a hash."""
        return hash(self._repr)


def _map_type_to_string(map_type: _basix.MapType) -> str:
    """Convert map type to a UFL string.

    Args:
        map_type: A map type.

    Returns:
        A string representing this map type.
    """
    if map_type == _basix.MapType.identity:
        return "identity"
    if map_type == _basix.MapType.L2Piola:
        return "L2 Piola"
    if map_type == _basix.MapType.contravariantPiola:
        return "contravariant Piola"
    if map_type == _basix.MapType.covariantPiola:
        return "covariant Piola"
    if map_type == _basix.MapType.doubleContravariantPiola:
        return "double contravariant Piola"
    if map_type == _basix.MapType.doubleCovariantPiola:
        return "double covariant Piola"
    raise ValueError(f"Unsupported map type: {map_type}")


def _compute_signature(element: _basix.finite_element.FiniteElement) -> str:
    """Compute a signature of a custom element.

    Args:
        element: A Basix custom element.

    Returns:
        A hash identifying this element.
    """
    assert element.family == _basix.ElementFamily.custom

    signature = (f"{element.cell_type.name}, {element.value_shape}, {element.map_type.name}, "
                 f"{element.discontinuous}, {element.highest_complete_degree}, {element.highest_degree}, ")
    data = ",".join([f"{i}" for row in element.wcoeffs for i in row])
    data += "__"
    for entity in element.x:
        for points in entity:
            data = ",".join([f"{i}" for p in points for i in p])
            data += "_"
    data += "__"
    for entity in element.M:
        for matrices in entity:
            data = ",".join([f"{i}" for mat in matrices for row in mat for i in row])
            data += "_"
    data += "__"
    for mat in element.entity_transformations().values():
        data = ",".join([f"{i}" for row in mat for i in row])
        data += "__"
    signature += _hashlib.sha1(data.encode('utf-8')).hexdigest()
    return signature


def create_element(family: _typing.Union[_basix.ElementFamily, str], cell: _typing.Union[_basix.CellType, str],
                   degree: int, lagrange_variant: _basix.LagrangeVariant = _basix.LagrangeVariant.unset,
                   dpc_variant: _basix.DPCVariant = _basix.DPCVariant.unset, discontinuous=False) -> BasixElement:
    """Create a UFL element using Basix.

    Args:
        family: The element's family as a Basix enum or a string.
        cell: The cell type as a Basix enum or a string.
        degree: The degree of the finite element.
        lagrange_variant: The variant of Lagrange to be used.
        dpc_variant: The variant of DPC to be used.
        discontinuous: If set to True, the discontinuous version of this element will be created.
    """
    if isinstance(cell, str):
        cell = _basix.cell.string_to_type(cell)
    if isinstance(family, str):
        if family.startswith("Discontinuous "):
            family = family[14:]
            discontinuous = True
        if family in ["DP", "DG", "DQ"]:
            family = "P"
            discontinuous = True
        if family == "DPC":
            discontinuous = True

        family = _basix.finite_element.string_to_family(family, cell.name)

    e = _basix.create_element(family, cell, degree, lagrange_variant, dpc_variant, discontinuous)
    return BasixElement(e)


def create_vector_element(
    family: _typing.Union[_basix.ElementFamily, str], cell: _typing.Union[_basix.CellType, str],
    degree: int, lagrange_variant: _basix.LagrangeVariant = _basix.LagrangeVariant.unset,
    dpc_variant: _basix.DPCVariant = _basix.DPCVariant.unset, discontinuous=False
) -> _ufl.VectorElement:
    """Create a UFL vector element using Basix.

    A vector element is an element which uses multiple copies of a scalar element to represent a
    vector-valued function.

    Args:
        family: The element's family as a Basix enum or a string.
        cell: The cell type as a Basix enum or a string.
        degree: The degree of the finite element.
        lagrange_variant: The variant of Lagrange to be used.
        dpc_variant: The variant of DPC to be used.
        discontinuous: If set to True, the discontinuous version of this element will be created.
    """
    e = create_element(family, cell, degree, lagrange_variant, dpc_variant, discontinuous)
    return _ufl.VectorElement(e)


def create_tensor_element(
    family: _typing.Union[_basix.ElementFamily, str], cell: _typing.Union[_basix.CellType, str],
    degree: int, lagrange_variant: _basix.LagrangeVariant = _basix.LagrangeVariant.unset,
    dpc_variant: _basix.DPCVariant = _basix.DPCVariant.unset, discontinuous=False
) -> _ufl.TensorElement:
    """Create a UFL tensor element using Basix.

    A tensor element is an element which uses multiple copies of a scalar element to represent a
    tensor-valued function.

    Args:
        family: The element's family as a Basix enum or a string.
        cell: The cell type as a Basix enum or a string.
        degree: The degree of the finite element.
        lagrange_variant: The variant of Lagrange to be used.
        dpc_variant: The variant of DPC to be used.
        discontinuous: If set to True, the discontinuous version of this element will be created.
    """
    e = create_element(family, cell, degree, lagrange_variant, dpc_variant, discontinuous)
    return _ufl.TensorElement(e)



# Copyright (c) 2020 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import pytest
import numpy as np
import random
from .utils import parametrize_over_elements


@parametrize_over_elements(3)
def test_if_permutations(cell_type, element_type, degree, element_args):
    e = basix.create_element(element_type, cell_type, degree, *element_args)

    for t in e.base_transformations():
        for row in t:
            a = np.argmax(row)
            if not np.isclose(row[a], 1) or not np.allclose(row[:a], 0) or not np.allclose(row[a + 1:], 0):
                assert not e.dof_transformations_are_permutations
                return
    assert e.dof_transformations_are_permutations


@parametrize_over_elements(3)
def test_if_identity(cell_type, element_type, degree, element_args):
    e = basix.create_element(element_type, cell_type, degree, *element_args)

    for t in e.base_transformations():
        if not np.allclose(t, np.eye(t.shape[0])):
            assert not e.dof_transformations_are_identity
            return
    assert e.dof_transformations_are_identity


@parametrize_over_elements(5)
def test_non_zero(cell_type, element_type, degree, element_args):
    e = basix.create_element(element_type, cell_type, degree, *element_args)
    for t in e.base_transformations():
        for row in t:
            assert max(abs(i) for i in row) > 1e-6


@parametrize_over_elements(5)
def test_apply_to_transpose(cell_type, element_type, degree, element_args):
    random.seed(42)

    e = basix.create_element(element_type, cell_type, degree, *element_args)

    size = e.dim

    for i in range(10):
        cell_info = random.randrange(2**30)

        data1 = np.array(list(range(size**2)), dtype=np.float32)
        data1 = e.apply_dof_transformation(data1, size, cell_info)
        data1 = data1.reshape((size, size))

        # This is the transpose of the data used above
        data2 = np.array([size * j + i for i in range(size) for j in range(size)], dtype=np.float32)
        data2 = e.apply_dof_transformation_to_transpose(data2, size, cell_info)
        data2 = data2.reshape((size, size))

        assert np.allclose(data1.transpose(), data2)


@parametrize_over_elements(5, basix.CellType.interval)
def test_interval_transformation_size(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.interval, degree, *element_args)
    assert len(e.base_transformations()) == 0


@parametrize_over_elements(5, basix.CellType.triangle)
def test_triangle_transformation_degrees(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.triangle, degree, *element_args)
    bt = e.base_transformations()
    assert len(bt) == 3
    identity = np.identity(e.dim)
    for i, degree in enumerate([2, 2, 2]):
        assert np.allclose(
            np.linalg.matrix_power(bt[i], degree),
            identity)


@parametrize_over_elements(5, basix.CellType.tetrahedron)
def test_tetrahedron_transformation_degrees(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.tetrahedron, degree, *element_args)
    bt = e.base_transformations()
    assert len(bt) == 14
    identity = np.identity(e.dim)
    for i, degree in enumerate([2, 2, 2, 2, 2, 2, 3, 2, 3, 2, 3, 2, 3, 2]):
        assert np.allclose(
            np.linalg.matrix_power(bt[i], degree),
            identity)


@parametrize_over_elements(5, basix.CellType.quadrilateral)
def test_quadrilateral_transformation_degrees(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.quadrilateral, degree, *element_args)
    bt = e.base_transformations()
    assert len(bt) == 4
    identity = np.identity(e.dim)
    for i, degree in enumerate([2, 2, 2, 2]):
        assert np.allclose(
            np.linalg.matrix_power(bt[i], degree),
            identity)


@parametrize_over_elements(5, basix.CellType.hexahedron)
def test_hexahedron_transformation_degrees(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.hexahedron, degree, *element_args)
    bt = e.base_transformations()
    assert len(bt) == 24
    identity = np.identity(e.dim)
    for i, degree in enumerate([2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                               4, 2, 4, 2, 4, 2, 4, 2, 4, 2, 4, 2]):
        # TODO: remove the atol here once non-equispaced variants are implemented
        assert np.allclose(
            np.linalg.matrix_power(bt[i], degree),
            identity, atol=1e-6)


@parametrize_over_elements(5, basix.CellType.prism)
def test_prism_transformation_degrees(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.prism, degree, *element_args)
    bt = e.base_transformations()
    assert len(bt) == 19
    identity = np.identity(e.dim)
    for i, degree in enumerate([2, 2, 2, 2, 2, 2, 2, 2, 2,
                               3, 2, 4, 2, 4, 2, 4, 2, 3, 2]):
        assert np.allclose(
            np.linalg.matrix_power(bt[i], degree),
            identity)


@parametrize_over_elements(5, basix.CellType.pyramid)
def test_pyramid_transformation_degrees(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.pyramid, degree, *element_args)
    bt = e.base_transformations()
    assert len(bt) == 18
    identity = np.identity(e.dim)
    for i, degree in enumerate([2, 2, 2, 2, 2, 2, 2, 2,
                               4, 2, 3, 2, 3, 2, 3, 2, 3, 2]):
        assert np.allclose(
            np.linalg.matrix_power(bt[i], degree),
            identity)


@parametrize_over_elements(5, basix.CellType.triangle)
def test_transformation_of_tabulated_data_triangle(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.triangle, degree, *element_args)
    bt = e.base_transformations()

    N = 4
    points = np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1 - i)])
    values = e.tabulate(0, points)[0]

    start = sum(e.num_entity_dofs[0])
    ndofs = e.num_entity_dofs[1][0]
    if ndofs != 0:
        # Check that the 0th transformation undoes the effect of reflecting edge 0
        reflected_points = np.array([[p[1], p[0]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[0, 1], [1, 0]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)

        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[0].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])


@parametrize_over_elements(5, basix.CellType.quadrilateral)
def test_transformation_of_tabulated_data_quadrilateral(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.quadrilateral, degree, *element_args)
    bt = e.base_transformations()

    N = 4
    points = np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1)])
    values = e.tabulate(0, points)[0]

    start = sum(e.num_entity_dofs[0])
    ndofs = e.num_entity_dofs[1][0]
    if ndofs != 0:
        # Check that the 0th transformation undoes the effect of reflecting edge 0
        reflected_points = np.array([[1 - p[0], p[1]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[-1, 0], [0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[0].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])


@parametrize_over_elements(5, basix.CellType.tetrahedron)
def test_transformation_of_tabulated_data_tetrahedron(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.tetrahedron, degree, *element_args)
    bt = e.base_transformations()

    N = 4
    points = np.array([[i / N, j / N, k / N]
                       for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1 - i - j)])
    values = e.tabulate(0, points)[0]

    start = sum(e.num_entity_dofs[0])
    ndofs = e.num_entity_dofs[1][0]
    if ndofs != 0:
        # Check that the 0th transformation undoes the effect of reflecting edge 0
        reflected_points = np.array([[p[0], p[2], p[1]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[0].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    start = sum(e.num_entity_dofs[0]) + sum(e.num_entity_dofs[1])
    ndofs = e.num_entity_dofs[2][0]
    if ndofs != 0:
        # Check that the 6th transformation undoes the effect of rotating face 0
        rotated_points = np.array([[p[2], p[0], p[1]] for p in points])
        rotated_values = e.tabulate(0, rotated_points)[0]

        _J = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(rotated_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose(bt[6].dot(i_slice)[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    if ndofs != 0:
        # Check that the 7th transformation undoes the effect of reflecting face 0
        reflected_points = np.array([[p[0], p[2], p[1]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[7].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])


@parametrize_over_elements(3, basix.CellType.hexahedron)
def test_transformation_of_tabulated_data_hexahedron(element_type, degree, element_args):
    if degree > 4 and element_type in [basix.ElementFamily.RT, basix.ElementFamily.N1E]:
        pytest.xfail("High degree Hdiv and Hcurl spaces on hexes based on "
                     "Lagrange spaces equally spaced points are unstable.")

    e = basix.create_element(element_type, basix.CellType.hexahedron, degree, *element_args)
    bt = e.base_transformations()

    N = 4
    points = np.array([[i / N, j / N, k / N]
                       for i in range(N + 1) for j in range(N + 1) for k in range(N + 1)])
    values = e.tabulate(0, points)[0]

    start = sum(e.num_entity_dofs[0])
    ndofs = e.num_entity_dofs[1][0]
    if ndofs != 0:
        # Check that the 0th transformation undoes the effect of reflecting edge 0
        reflected_points = np.array([[1 - p[0], p[1], p[2]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[0].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    start = sum(e.num_entity_dofs[0]) + sum(e.num_entity_dofs[1])
    ndofs = e.num_entity_dofs[2][0]
    if ndofs != 0:
        # Check that the 12th transformation undoes the effect of rotating face 0
        rotated_points = np.array([[1 - p[1], p[0], p[2]] for p in points])
        rotated_values = e.tabulate(0, rotated_points)[0]

        _J = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(rotated_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose(bt[12].dot(i_slice)[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    if ndofs != 0:
        # Check that the 13th transformation undoes the effect of reflecting face 0
        reflected_points = np.array([[p[1], p[0], p[2]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[13].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])


@parametrize_over_elements(3, basix.CellType.prism)
def test_transformation_of_tabulated_data_prism(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.prism, degree, *element_args)
    bt = e.base_transformations()

    N = 4
    points = np.array([[i / N, j / N, k / N]
                       for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1)])
    values = e.tabulate(0, points)[0]

    start = sum(e.num_entity_dofs[0])
    ndofs = e.num_entity_dofs[1][0]
    if ndofs != 0:
        # Check that the 0th transformation undoes the effect of reflecting edge 0
        reflected_points = np.array([[1 - p[1] - p[0], p[1], p[2]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[-1, 0, 0], [-1, 1, 0], [0, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[0].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    start = sum(e.num_entity_dofs[0]) + sum(e.num_entity_dofs[1])
    ndofs = e.num_entity_dofs[2][0]
    if ndofs != 0:
        # Check that the 10th transformation undoes the effect of rotating face 0
        rotated_points = np.array([[1 - p[0] - p[1], p[0], p[2]] for p in points])
        rotated_values = e.tabulate(0, rotated_points)[0]

        _J = np.array([[-1, 1, 0], [-1, 0, 0], [0, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(rotated_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose(bt[10].dot(i_slice)[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    if ndofs != 0:
        # Check that the 11th transformation undoes the effect of reflecting face 0
        reflected_points = np.array([[p[1], p[0], p[2]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[11].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])


@parametrize_over_elements(3, basix.CellType.pyramid)
def test_transformation_of_tabulated_data_pyramid(element_type, degree, element_args):
    e = basix.create_element(element_type, basix.CellType.pyramid, degree, *element_args)
    bt = e.base_transformations()

    N = 4
    points = np.array([[i / N, j / N, k / N]
                       for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1)])
    values = e.tabulate(0, points)[0]

    start = sum(e.num_entity_dofs[0])
    ndofs = e.num_entity_dofs[1][0]
    if ndofs != 0:
        # Check that the 0th transformation undoes the effect of reflecting edge 0
        reflected_points = np.array([[1 - p[2] - p[0], p[1], p[2]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[-1, 0, 0], [0, 1, 0], [-1, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[0].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    start = sum(e.num_entity_dofs[0]) + sum(e.num_entity_dofs[1])
    ndofs = e.num_entity_dofs[2][0]
    if ndofs != 0:
        # Check that the 8th transformation undoes the effect of rotating face 0
        rotated_points = np.array([[1 - p[1] - p[2], p[0], p[2]] for p in points])
        rotated_values = e.tabulate(0, rotated_points)[0]

        _J = np.array([[0, 1, 0], [-1, 0, 0], [-1, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(rotated_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose(bt[8].dot(i_slice)[start: start + ndofs],
                                   j_slice[start: start + ndofs])

    if ndofs != 0:
        # Check that the 9th transformation undoes the effect of reflecting face 0
        reflected_points = np.array([[p[1], p[0], p[2]] for p in points])
        reflected_values = e.tabulate(0, reflected_points)[0]

        _J = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]])
        J = np.array([_J for p in points])
        detJ = np.array([np.linalg.det(_J) for p in points])
        K = np.array([np.linalg.inv(_J) for p in points])
        mapped_values = e.push_forward(reflected_values, J, detJ, K)
        for i, j in zip(values, mapped_values):
            for d in range(e.value_size):
                i_slice = i[:, d]
                j_slice = j[:, d]
                assert np.allclose((bt[9].dot(i_slice))[start: start + ndofs],
                                   j_slice[start: start + ndofs])



import os
import pylit

pylit.defaults.text_extensions = [".rst"]

path = os.path.dirname(os.path.realpath(__file__))
for file in os.listdir(path):
    if file.endswith(".py") and file.startswith("demo"):
        pylit.main([file])



# ==========================
# Computing a facet integral
# ==========================
#
# In this demo, we look at how Basix can be used to compute the integral
# of the normal derivative of a basis function over a triangular facet
# of a tetrahedral cell.
#
# As an example, we integrate the normal derivative of the fifth basis
# function (note: counting starts at 0) of a degree 3 Lagrange space
# over the zeroth facet of a tetrahedral cell. This facet will have
# vertices at (1,0,0), (0,1,0) and (0,0,1).
#
# We start by importing Basis and Numpy.

import basix
import numpy as np
from basix import ElementFamily, CellType, LagrangeVariant

# We define a degree 3 Lagrange space on a tetrahedron.

lagrange = basix.create_element(
    ElementFamily.P, CellType.tetrahedron, 3, LagrangeVariant.equispaced)

# The facets of a tetrahedron are triangular, so we create a quadrature
# rule on a triangle. We use an order 3 rule so that we can integrate the
# basis functions in our space exactly.

points, weights = basix.make_quadrature(CellType.triangle, 3)

# Next, we must map the quadrature points to our facet. We use the function
# `geometry` to get the coordinates of the vertices of the tetrahedron, and
# we use `sub_entity_connectivity` to see which vertices are adjacent to
# our facet. We get the sub-entity connectivity using the indices 2 (facets
# of 3D cells have dimension 2), 0 (vertices have dimension 0), and 0 (the
# index of the facet we chose to use).
#
# Using this information, we can map the quadrature points to the facet.

vertices = basix.geometry(CellType.tetrahedron)
facet = basix.cell.sub_entity_connectivity(CellType.tetrahedron)[2][0][0]
mapped_points = np.array([
    vertices[facet[0]] * (1 - x - y) + vertices[facet[1]] * x + vertices[facet[2]] * y
    for x, y in points
])

# We now compute the normal derivative of the fifth basis function at
# the quadrature points. First, we use `facet_outward_normals` to get
# the normal vector to the facet.
#
# We then tabulate the basis functions of our space at the quadrature
# points. We pass 1 in as the first argument, as we want the derivatives
# of the basis functions. The result of tabulation will be an array of
# size 4 by number of quadrature points by number of degrees of freedom.
# To get the data that we want, we use the indices `1:` (to get the
# derivatives and not also the function values), `:` (to include every
# point), 5 (to get the fifth basis function), and 0 (to get the only
# entry as the value size is 1).
#
# We then multiply the three derivatives of the basis function by
# the three componenets of the normal.

normal = basix.cell.facet_outward_normals(CellType.tetrahedron)[0]
tab = lagrange.tabulate(1, mapped_points)[1:, :, 5, 0]
normal_deriv = tab[0] * normal[0] + tab[1] * normal[1] + tab[2] * normal[2]

# As our facet is not the reference triangle, we must multiply the
# integrand by the norm of the Jacobian. We compute this by taking the
# cross product of the two columns of the Jacobian, and then compute
# the integral.

jacobian = basix.cell.facet_jacobians(CellType.tetrahedron)[0]
size_jacobian = np.linalg.norm(np.cross(jacobian[:, 0], jacobian[:, 1]))
print(np.sum(normal_deriv * weights) * size_jacobian)



# Copyright (c) 2020 Matthew Scroggs
# FEniCS Project
# SPDX-License-Identifier: MIT

import numpy as np
import basix
import pytest
from .utils import parametrize_over_elements


@pytest.mark.parametrize("n", range(1, 6))
@pytest.mark.parametrize("cell_type", [basix.CellType.interval, basix.CellType.triangle, basix.CellType.tetrahedron])
@pytest.mark.parametrize("element_type", [basix.ElementFamily.P])
def test_interpolation(cell_type, n, element_type):
    element = basix.create_element(element_type, cell_type, n, basix.LagrangeVariant.gll_warped)
    assert element.interpolation_matrix.shape[0] == element.dim
    assert element.interpolation_matrix.shape[1] == element.points.shape[0]
    assert element.points.shape[1] == len(basix.topology(element.cell_type)) - 1


@parametrize_over_elements(5)
def test_interpolation_matrix(cell_type, degree, element_type, element_args):
    if degree > 4:
        if cell_type in [
            basix.CellType.quadrilateral, basix.CellType.hexahedron
        ] and element_type in [
            basix.ElementFamily.RT, basix.ElementFamily.N1E, basix.ElementFamily.BDM,
            basix.ElementFamily.N2E
        ]:
            pytest.xfail("High degree Hdiv and Hcurl spaces on hexes based on "
                         "Lagrange spaces with equally spaced points are unstable.")

    element = basix.create_element(element_type, cell_type, degree, *element_args)
    i_m = element.interpolation_matrix
    tabulated = element.tabulate(0, element.points)[0]

    # Loop over dofs
    coeffs = np.zeros((i_m.shape[0], i_m.shape[0]))
    for i in range(i_m.shape[0]):
        coeffs[i, :] = i_m @ tabulated[:, i::i_m.shape[0]].T.reshape(i_m.shape[1])

    assert np.allclose(coeffs, np.identity(coeffs.shape[0]))


@parametrize_over_elements(4)
def test_interpolation_is_identity(cell_type, degree, element_type, element_args):
    element = basix.create_element(element_type, cell_type, degree, *element_args)
    i_m = element.interpolation_matrix

    if i_m.shape[0] == i_m.shape[1]:
        assert element.interpolation_is_identity == np.allclose(i_m, np.eye(i_m.shape[0]))
    else:
        assert not element.interpolation_is_identity



# Copyright (c) 2020 Chris Richardson
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import numpy as np
import pytest
import sympy


@pytest.mark.parametrize("celltype", [(basix.CellType.quadrilateral, 1.0),
                                      (basix.CellType.hexahedron, 1.0),
                                      (basix.CellType.prism, 0.5),
                                      (basix.CellType.interval, 1.0),
                                      (basix.CellType.triangle, 0.5),
                                      (basix.CellType.tetrahedron, 1.0/6.0)])
@pytest.mark.parametrize("order", range(9))
def test_cell_quadrature(celltype, order):
    Qpts, Qwts = basix.make_quadrature(celltype[0], order)
    assert(np.isclose(sum(Qwts), celltype[1]))


@pytest.mark.parametrize("m", range(7))
@pytest.mark.parametrize("scheme", [basix.QuadratureType.Default, basix.QuadratureType.gll])
def test_qorder_line(m, scheme):
    Qpts, Qwts = basix.make_quadrature(scheme, basix.CellType.interval, m)
    x = sympy.Symbol('x')
    f = x**m
    q = sympy.integrate(f, (x, 0, (1)))
    s = 0.0
    for (pt, wt) in zip(Qpts, Qwts):
        s += wt * f.subs([(x, pt[0])])
    assert(np.isclose(float(q), float(s)))


@pytest.mark.parametrize("m", range(6))
@pytest.mark.parametrize("scheme", [basix.QuadratureType.Default, basix.QuadratureType.gauss_jacobi])
def test_qorder_tri(m, scheme):
    Qpts, Qwts = basix.make_quadrature(scheme, basix.CellType.triangle, m)
    x = sympy.Symbol('x')
    y = sympy.Symbol('y')
    f = x**m + y**m
    q = sympy.integrate(sympy.integrate(f, (x, 0, (1 - y))), (y, 0, 1))
    s = 0.0
    for (pt, wt) in zip(Qpts, Qwts):
        s += wt * f.subs([(x, pt[0]), (y, pt[1])])
    assert(np.isclose(float(q), float(s)))


@pytest.mark.parametrize("m", range(1, 20))
@pytest.mark.parametrize("scheme", [basix.QuadratureType.xiao_gimbutas])
def test_xiao_gimbutas_tri(m, scheme):
    Qpts, Qwts = basix.make_quadrature(scheme, basix.CellType.triangle, m)
    x = sympy.Symbol('x')
    y = sympy.Symbol('y')
    f = x**m + 2 * y**m
    q = sympy.integrate(sympy.integrate(f, (x, 0, 1 - y)), (y, 0, 1))
    s = 0.0
    for (pt, wt) in zip(Qpts, Qwts):
        s += wt * f.subs([(x, pt[0]), (y, pt[1])])
    assert(np.isclose(float(q), float(s)))


@pytest.mark.parametrize("m", range(1, 16))
@pytest.mark.parametrize("scheme", [basix.QuadratureType.xiao_gimbutas])
def test_xiao_gimbutas_tet(m, scheme):
    Qpts, Qwts = basix.make_quadrature(scheme, basix.CellType.tetrahedron, m)
    x = sympy.Symbol('x')
    y = sympy.Symbol('y')
    z = sympy.Symbol('z')
    f = x**m + 2 * y**m + 3 * z**m
    q = sympy.integrate(f, (x, 0, 1 - y - z), (y, 0, 1 - z), (z, 0, 1))
    s = 0.0
    for (pt, wt) in zip(Qpts, Qwts):
        s += wt * f.subs([(x, pt[0]), (y, pt[1]), (z, pt[2])])
    assert(np.isclose(float(q), float(s)))


@pytest.mark.parametrize("m", range(9))
@pytest.mark.parametrize("scheme", [basix.QuadratureType.Default, basix.QuadratureType.gauss_jacobi])
def test_qorder_tet(m, scheme):
    Qpts, Qwts = basix.make_quadrature(scheme, basix.CellType.tetrahedron, m)
    x = sympy.Symbol('x')
    y = sympy.Symbol('y')
    z = sympy.Symbol('z')
    f = x**m + y**m + z**m
    q = sympy.integrate(sympy.integrate(sympy.integrate(f, (x, 0, (1 - y - z))), (y, 0, 1 - z)), (z, 0, 1))
    s = 0.0
    for (pt, wt) in zip(Qpts, Qwts):
        s += wt * f.subs([(x, pt[0]), (y, pt[1]), (z, pt[2])])
    assert(np.isclose(float(q), float(s)))


def test_quadrature_function():
    Qpts, Qwts = basix.make_quadrature(basix.CellType.interval, 3)
    # Scale to interval [0.0, 2.0]
    Qpts *= 2.0
    Qwts *= 2.0

    def f(x):
        return x * x

    b = sum([w * f(pt[0]) for pt, w in zip(Qpts, Qwts)])

    assert np.isclose(b, 8.0 / 3.0)


def test_gll():
    m = 5

    # 1D interval
    pts, wts = basix.make_quadrature(basix.QuadratureType.gll, basix.CellType.interval, m+1)
    pts, wts = 2*pts.flatten()-1, 2*wts.flatten()
    ref_pts = np.array([-1., 1., -np.sqrt(3/7), 0.0, np.sqrt(3/7)])
    assert (np.allclose(pts.flatten(), ref_pts))
    ref_wts = np.array([1/10, 1/10, 49/90, 32/45, 49/90])
    assert (np.allclose(wts, ref_wts))
    assert np.isclose(sum(pts * wts), 0)
    assert np.isclose(sum(wts), 2)

    # 2D quad
    pts, wts = basix.make_quadrature(basix.QuadratureType.gll, basix.CellType.quadrilateral, m+1)
    pts, wts = 2*pts-1, 4*wts
    ref_pts2 = np.array([[x, y] for x in ref_pts for y in ref_pts])
    assert (np.allclose(pts, ref_pts2))
    ref_wts2 = np.array([w1*w2 for w1 in ref_wts for w2 in ref_wts])
    assert (np.allclose(wts, ref_wts2))
    assert np.isclose((pts * wts.reshape(-1, 1)).sum(), 0)
    assert np.isclose(sum(wts), 4)

    # 3D hex
    pts, wts = basix.make_quadrature(basix.QuadratureType.gll, basix.CellType.hexahedron, m+1)
    pts, wts = 2*pts-1, 8*wts
    ref_pts3 = np.array([[x, y, z] for x in ref_pts for y in ref_pts for z in ref_pts])
    assert (np.allclose(pts, ref_pts3))
    ref_wts3 = np.array([w1*w2*w3 for w1 in ref_wts for w2 in ref_wts for w3 in ref_wts])
    assert (np.allclose(wts, ref_wts3))
    assert np.isclose((pts * wts.reshape(-1, 1)).sum(), 0)
    assert np.isclose(sum(wts), 8)



# ==============================================
# Defining conforming Crouzeix--Raviart elements
# ==============================================
#
# In this demo, we show how Basix's custom element functionality can be used to
# create a conforming Crouzeix--Raviart element.
#
# First, we import Basix and Numpy.

import basix
import numpy as np
from basix import CellType, MapType, PolynomialType, LatticeType

from mpl_toolkits import mplot3d  # noqa: F401
import matplotlib.pyplot as plt

# Conforming CR element on a triangle
# ===================================
#
# We begin by implementing this element on a triangle. The following function
# implements this element for an arbitrary degree. Details of the definition of
# this element can be found at
# https://defelement.com/elements/conforming-crouzeix-raviart.html.
#
# As the input to this function, we use the degree of the element as shown on
# DefElement. For most degrees, the highest degree polynomial in this elements
# polynomial space is actually one degree higher, so we pass `degree + 1` into
# `create_custom_element`.


def create_ccr_triangle(degree):

    if degree == 1:
        wcoeffs = np.eye(3)

        x = [[], [], [], []]
        x[0].append(np.array([[0.0, 0.0]]))
        x[0].append(np.array([[1.0, 0.0]]))
        x[0].append(np.array([[0.0, 1.0]]))

        for _ in range(3):
            x[1].append(np.zeros((0, 2)))
        x[2].append(np.zeros((0, 2)))

        M = [[], [], [], []]
        for _ in range(3):
            M[0].append(np.array([[[1.]]]))

        for _ in range(3):
            M[1].append(np.zeros((0, 1, 0)))
        M[2].append(np.zeros((0, 1, 0)))

        return basix.create_custom_element(
            CellType.triangle, [], wcoeffs, x, M, MapType.identity, False, 1, 1)

    npoly = (degree + 2) * (degree + 3) // 2
    ndofs = degree * (degree + 5) // 2
    wcoeffs = np.zeros((ndofs, npoly))

    dof_n = 0
    for i in range((degree + 1) * (degree + 2) // 2):
        wcoeffs[dof_n, dof_n] = 1
        dof_n += 1

    pts, wts = basix.make_quadrature(CellType.triangle, 2 * (degree + 1))
    poly = basix.tabulate_polynomials(PolynomialType.legendre, CellType.triangle, degree + 1, pts)
    for i in range(1, degree):
        x = pts[:, 0]
        y = pts[:, 1]
        f = x ** i * y ** (degree - i) * (x + y)

        for j in range(npoly):
            wcoeffs[dof_n, j] = sum(f * poly[:, j] * wts)
        dof_n += 1

    geometry = basix.geometry(CellType.triangle)
    topology = basix.topology(CellType.triangle)
    x = [[], [], [], []]
    M = [[], [], [], []]
    for v in topology[0]:
        x[0].append(np.array(geometry[v]))
        M[0].append(np.array([[[1.]]]))
    pts = basix.create_lattice(CellType.interval, degree, LatticeType.equispaced, False)
    mat = np.zeros((len(pts), 1, len(pts)))
    mat[:, 0, :] = np.eye(len(pts))
    for e in topology[1]:
        edge_pts = []
        v0 = geometry[e[0]]
        v1 = geometry[e[1]]
        for p in pts:
            edge_pts.append(v0 + p * (v1 - v0))
        x[1].append(np.array(edge_pts))
        M[1].append(mat)
    pts = basix.create_lattice(CellType.triangle, degree + 1, LatticeType.equispaced, False)
    x[2].append(pts)
    mat = np.zeros((len(pts), 1, len(pts)))
    mat[:, 0, :] = np.eye(len(pts))
    M[2].append(mat)

    return basix.create_custom_element(
        CellType.triangle, [], wcoeffs, x, M, MapType.identity, False, degree, degree + 1)


# We can then create a degree 2 conforming CR element.

e = create_ccr_triangle(2)

# We now visualise the basis functions of the element we have created.

pts = basix.create_lattice(CellType.triangle, 30, LatticeType.equispaced, True)

x = pts[:, 0]
y = pts[:, 1]
z = e.tabulate(0, pts)[0]

fig = plt.figure(figsize=(8, 8))

for n in range(7):
    if n == 6:
        ax = plt.subplot(3, 3, n + 2, projection='3d')
    else:
        ax = plt.subplot(3, 3, n + 1, projection='3d')
    ax.plot([0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0], "k-")
    ax.scatter(x, y, z[:, n, 0])

plt.savefig("ccr_triangle_2.png")

# .. image:: ccr_triangle_2.png
#  :width: 100%
#  :alt: The basis functions of a degree 2 conforming CR element

# We also visualise the basis functions of a degree 3 conforming CR element.

e = create_ccr_triangle(3)

pts = basix.create_lattice(CellType.triangle, 30, LatticeType.equispaced, True)

x = pts[:, 0]
y = pts[:, 1]
z = e.tabulate(0, pts)[0]

fig = plt.figure(figsize=(11, 8))

for n in range(12):
    ax = plt.subplot(3, 4, n + 1, projection='3d')
    ax.plot([0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0], "k-")
    ax.scatter(x, y, z[:, n, 0])

plt.savefig("ccr_triangle_3.png")

# .. image:: ccr_triangle_3.png
#  :width: 100%
#  :alt: The basis functions of a degree 3 conforming CR element



# Copyright (c) 2020 Chris Richardson
# FEniCS Project
# SPDX-License-Identifier: MIT

import basix
import numpy
import pytest
import sympy


def sympy_lagrange(celltype, n):
    # These basis functions were computed using symfem. They can be recomputed
    # by running (eg):
    #    import symfem
    #    e = symfem.create_element("triangle", "Lagrange", 2)
    #    print(e.get_basis_functions())
    x = sympy.Symbol("x")
    y = sympy.Symbol("y")
    z = sympy.Symbol("z")

    if celltype == basix.CellType.interval:
        if n == 1:
            return [1 - x, x]
        if n == 2:
            return [2*x**2 - 3*x + 1, 2*x**2 - x, -4*x**2 + 4*x]
        if n == 3:
            return [-9*x**3/2 + 9*x**2 - 11*x/2 + 1,
                    9*x**3/2 - 9*x**2/2 + x,
                    27*x**3/2 - 45*x**2/2 + 9*x,
                    -27*x**3/2 + 18*x**2 - 9*x/2]
        if n == 4:
            return [32*x**4/3 - 80*x**3/3 + 70*x**2/3 - 25*x/3 + 1,
                    32*x**4/3 - 16*x**3 + 22*x**2/3 - x,
                    -128*x**4/3 + 96*x**3 - 208*x**2/3 + 16*x,
                    64*x**4 - 128*x**3 + 76*x**2 - 12*x,
                    -128*x**4/3 + 224*x**3/3 - 112*x**2/3 + 16*x/3]
        if n == 5:
            return [-625*x**5/24 + 625*x**4/8 - 2125*x**3/24 + 375*x**2/8 - 137*x/12 + 1,
                    625*x**5/24 - 625*x**4/12 + 875*x**3/24 - 125*x**2/12 + x,
                    3125*x**5/24 - 4375*x**4/12 + 8875*x**3/24 - 1925*x**2/12 + 25*x,
                    -3125*x**5/12 + 8125*x**4/12 - 7375*x**3/12 + 2675*x**2/12 - 25*x,
                    3125*x**5/12 - 625*x**4 + 6125*x**3/12 - 325*x**2/2 + 50*x/3,
                    -3125*x**5/24 + 6875*x**4/24 - 5125*x**3/24 + 1525*x**2/24 - 25*x/4]
    if celltype == basix.CellType.triangle:
        if n == 1:
            return [-x - y + 1, x, y]
        if n == 2:
            return [2*x**2 + 4*x*y - 3*x + 2*y**2 - 3*y + 1,
                    2*x**2 - x,
                    2*y**2 - y,
                    4*x*y,
                    -4*x*y - 4*y**2 + 4*y,
                    -4*x**2 - 4*x*y + 4*x]
        if n == 3:
            return [
                -9*x**3/2 - 27*x**2*y/2 + 9*x**2 - 27*x*y**2/2 + 18*x*y - 11*x/2 - 9*y**3/2 + 9*y**2 - 11*y/2 + 1,
                9*x**3/2 - 9*x**2/2 + x,
                9*y**3/2 - 9*y**2/2 + y,
                27*x**2*y/2 - 9*x*y/2,
                27*x*y**2/2 - 9*x*y/2,
                27*x**2*y/2 + 27*x*y**2 - 45*x*y/2 + 27*y**3/2 - 45*y**2/2 + 9*y,
                -27*x*y**2/2 + 9*x*y/2 - 27*y**3/2 + 18*y**2 - 9*y/2,
                27*x**3/2 + 27*x**2*y - 45*x**2/2 + 27*x*y**2/2 - 45*x*y/2 + 9*x,
                -27*x**3/2 - 27*x**2*y/2 + 18*x**2 + 9*x*y/2 - 9*x/2,
                -27*x**2*y - 27*x*y**2 + 27*x*y
            ]
        if n == 4:
            return [
                32*x**4/3 + 128*x**3*y/3 - 80*x**3/3 + 64*x**2*y**2 - 80*x**2*y + 70*x**2/3 + 128*x*y**3/3 - 80*x*y**2 + 140*x*y/3 - 25*x/3 + 32*y**4/3 - 80*y**3/3 + 70*y**2/3 - 25*y/3 + 1,  # noqa: E501
                32*x**4/3 - 16*x**3 + 22*x**2/3 - x,
                32*y**4/3 - 16*y**3 + 22*y**2/3 - y,
                128*x**3*y/3 - 32*x**2*y + 16*x*y/3,
                64*x**2*y**2 - 16*x**2*y - 16*x*y**2 + 4*x*y,
                128*x*y**3/3 - 32*x*y**2 + 16*x*y/3,
                -128*x**3*y/3 - 128*x**2*y**2 + 96*x**2*y - 128*x*y**3 + 192*x*y**2 - 208*x*y/3 - 128*y**4/3 + 96*y**3 - 208*y**2/3 + 16*y,  # noqa: E501
                64*x**2*y**2 - 16*x**2*y + 128*x*y**3 - 144*x*y**2 + 28*x*y + 64*y**4 - 128*y**3 + 76*y**2 - 12*y,
                -128*x*y**3/3 + 32*x*y**2 - 16*x*y/3 - 128*y**4/3 + 224*y**3/3 - 112*y**2/3 + 16*y/3,
                -128*x**4/3 - 128*x**3*y + 96*x**3 - 128*x**2*y**2 + 192*x**2*y - 208*x**2/3 - 128*x*y**3/3 + 96*x*y**2 - 208*x*y/3 + 16*x,  # noqa: E501
                64*x**4 + 128*x**3*y - 128*x**3 + 64*x**2*y**2 - 144*x**2*y + 76*x**2 - 16*x*y**2 + 28*x*y - 12*x,
                -128*x**4/3 - 128*x**3*y/3 + 224*x**3/3 + 32*x**2*y - 112*x**2/3 - 16*x*y/3 + 16*x/3,
                128*x**3*y + 256*x**2*y**2 - 224*x**2*y + 128*x*y**3 - 224*x*y**2 + 96*x*y,
                -128*x**3*y - 128*x**2*y**2 + 160*x**2*y + 32*x*y**2 - 32*x*y,
                -128*x**2*y**2 + 32*x**2*y - 128*x*y**3 + 160*x*y**2 - 32*x*y
            ]
        if n == 5:
            return [
                -625*x**5/24 - 3125*x**4*y/24 + 625*x**4/8 - 3125*x**3*y**2/12 + 625*x**3*y/2 - 2125*x**3/24 - 3125*x**2*y**3/12 + 1875*x**2*y**2/4 - 2125*x**2*y/8 + 375*x**2/8 - 3125*x*y**4/24 + 625*x*y**3/2 - 2125*x*y**2/8 + 375*x*y/4 - 137*x/12 - 625*y**5/24 + 625*y**4/8 - 2125*y**3/24 + 375*y**2/8 - 137*y/12 + 1,  # noqa: E501
                625*x**5/24 - 625*x**4/12 + 875*x**3/24 - 125*x**2/12 + x,
                625*y**5/24 - 625*y**4/12 + 875*y**3/24 - 125*y**2/12 + y,
                3125*x**4*y/24 - 625*x**3*y/4 + 1375*x**2*y/24 - 25*x*y/4,
                3125*x**3*y**2/12 - 625*x**3*y/12 - 625*x**2*y**2/4 + 125*x**2*y/4 + 125*x*y**2/6 - 25*x*y/6,
                3125*x**2*y**3/12 - 625*x**2*y**2/4 + 125*x**2*y/6 - 625*x*y**3/12 + 125*x*y**2/4 - 25*x*y/6,
                3125*x*y**4/24 - 625*x*y**3/4 + 1375*x*y**2/24 - 25*x*y/4,
                3125*x**4*y/24 + 3125*x**3*y**2/6 - 4375*x**3*y/12 + 3125*x**2*y**3/4 - 4375*x**2*y**2/4 + 8875*x**2*y/24 + 3125*x*y**4/6 - 4375*x*y**3/4 + 8875*x*y**2/12 - 1925*x*y/12 + 3125*y**5/24 - 4375*y**4/12 + 8875*y**3/24 - 1925*y**2/12 + 25*y,  # noqa: E501
                -3125*x**3*y**2/12 + 625*x**3*y/12 - 3125*x**2*y**3/4 + 3125*x**2*y**2/4 - 125*x**2*y - 3125*x*y**4/4 + 5625*x*y**3/4 - 8875*x*y**2/12 + 1175*x*y/12 - 3125*y**5/12 + 8125*y**4/12 - 7375*y**3/12 + 2675*y**2/12 - 25*y,  # noqa: E501
                3125*x**2*y**3/12 - 625*x**2*y**2/4 + 125*x**2*y/6 + 3125*x*y**4/6 - 3125*x*y**3/4 + 3875*x*y**2/12 - 75*x*y/2 + 3125*y**5/12 - 625*y**4 + 6125*y**3/12 - 325*y**2/2 + 50*y/3,  # noqa: E501
                -3125*x*y**4/24 + 625*x*y**3/4 - 1375*x*y**2/24 + 25*x*y/4 - 3125*y**5/24 + 6875*y**4/24 - 5125*y**3/24 + 1525*y**2/24 - 25*y/4,  # noqa: E501
                3125*x**5/24 + 3125*x**4*y/6 - 4375*x**4/12 + 3125*x**3*y**2/4 - 4375*x**3*y/4 + 8875*x**3/24 + 3125*x**2*y**3/6 - 4375*x**2*y**2/4 + 8875*x**2*y/12 - 1925*x**2/12 + 3125*x*y**4/24 - 4375*x*y**3/12 + 8875*x*y**2/24 - 1925*x*y/12 + 25*x,  # noqa: E501
                -3125*x**5/12 - 3125*x**4*y/4 + 8125*x**4/12 - 3125*x**3*y**2/4 + 5625*x**3*y/4 - 7375*x**3/12 - 3125*x**2*y**3/12 + 3125*x**2*y**2/4 - 8875*x**2*y/12 + 2675*x**2/12 + 625*x*y**3/12 - 125*x*y**2 + 1175*x*y/12 - 25*x,  # noqa: E501
                3125*x**5/12 + 3125*x**4*y/6 - 625*x**4 + 3125*x**3*y**2/12 - 3125*x**3*y/4 + 6125*x**3/12 - 625*x**2*y**2/4 + 3875*x**2*y/12 - 325*x**2/2 + 125*x*y**2/6 - 75*x*y/2 + 50*x/3,  # noqa: E501
                -3125*x**5/24 - 3125*x**4*y/24 + 6875*x**4/24 + 625*x**3*y/4 - 5125*x**3/24 - 1375*x**2*y/24 + 1525*x**2/24 + 25*x*y/4 - 25*x/4,  # noqa: E501
                -3125*x**4*y/6 - 3125*x**3*y**2/2 + 1250*x**3*y - 3125*x**2*y**3/2 + 2500*x**2*y**2 - 5875*x**2*y/6 - 3125*x*y**4/6 + 1250*x*y**3 - 5875*x*y**2/6 + 250*x*y,  # noqa: E501
                3125*x**4*y/4 + 3125*x**3*y**2/2 - 3125*x**3*y/2 + 3125*x**2*y**3/4 - 6875*x**2*y**2/4 + 3625*x**2*y/4 - 625*x*y**3/4 + 1125*x*y**2/4 - 125*x*y,  # noqa: E501
                -3125*x**4*y/6 - 3125*x**3*y**2/6 + 2500*x**3*y/3 + 625*x**2*y**2/2 - 2125*x**2*y/6 - 125*x*y**2/3 + 125*x*y/3,  # noqa: E501
                3125*x**3*y**2/4 - 625*x**3*y/4 + 3125*x**2*y**3/2 - 6875*x**2*y**2/4 + 1125*x**2*y/4 + 3125*x*y**4/4 - 3125*x*y**3/2 + 3625*x*y**2/4 - 125*x*y,  # noqa: E501
                -3125*x**3*y**2/4 + 625*x**3*y/4 - 3125*x**2*y**3/4 + 4375*x**2*y**2/4 - 375*x**2*y/2 + 625*x*y**3/4 - 375*x*y**2/2 + 125*x*y/4,  # noqa: E501
                -3125*x**2*y**3/6 + 625*x**2*y**2/2 - 125*x**2*y/3 - 3125*x*y**4/6 + 2500*x*y**3/3 - 2125*x*y**2/6 + 125*x*y/3  # noqa: E501
            ]
    if celltype == basix.CellType.tetrahedron:
        if n == 1:
            return [-x - y - z + 1, x, y, z]
        if n == 2:
            return [2*x**2 + 4*x*y + 4*x*z - 3*x + 2*y**2 + 4*y*z - 3*y + 2*z**2 - 3*z + 1,
                    2*x**2 - x,
                    2*y**2 - y,
                    2*z**2 - z,
                    4*y*z,
                    4*x*z,
                    4*x*y,
                    -4*x*z - 4*y*z - 4*z**2 + 4*z,
                    -4*x*y - 4*y**2 - 4*y*z + 4*y,
                    -4*x**2 - 4*x*y - 4*x*z + 4*x]
        if n == 3:
            return [
                -9*x**3/2 - 27*x**2*y/2 - 27*x**2*z/2 + 9*x**2 - 27*x*y**2/2 - 27*x*y*z + 18*x*y - 27*x*z**2/2 + 18*x*z - 11*x/2 - 9*y**3/2 - 27*y**2*z/2 + 9*y**2 - 27*y*z**2/2 + 18*y*z - 11*y/2 - 9*z**3/2 + 9*z**2 - 11*z/2 + 1,  # noqa: E501
                9*x**3/2 - 9*x**2/2 + x,
                9*y**3/2 - 9*y**2/2 + y,
                9*z**3/2 - 9*z**2/2 + z,
                27*y**2*z/2 - 9*y*z/2,
                27*y*z**2/2 - 9*y*z/2,
                27*x**2*z/2 - 9*x*z/2,
                27*x*z**2/2 - 9*x*z/2,
                27*x**2*y/2 - 9*x*y/2,
                27*x*y**2/2 - 9*x*y/2,
                27*x**2*z/2 + 27*x*y*z + 27*x*z**2 - 45*x*z/2 + 27*y**2*z/2 + 27*y*z**2 - 45*y*z/2 + 27*z**3/2 - 45*z**2/2 + 9*z,  # noqa: E501
                -27*x*z**2/2 + 9*x*z/2 - 27*y*z**2/2 + 9*y*z/2 - 27*z**3/2 + 18*z**2 - 9*z/2,
                27*x**2*y/2 + 27*x*y**2 + 27*x*y*z - 45*x*y/2 + 27*y**3/2 + 27*y**2*z - 45*y**2/2 + 27*y*z**2/2 - 45*y*z/2 + 9*y,  # noqa: E501
                -27*x*y**2/2 + 9*x*y/2 - 27*y**3/2 - 27*y**2*z/2 + 18*y**2 + 9*y*z/2 - 9*y/2,
                27*x**3/2 + 27*x**2*y + 27*x**2*z - 45*x**2/2 + 27*x*y**2/2 + 27*x*y*z - 45*x*y/2 + 27*x*z**2/2 - 45*x*z/2 + 9*x,  # noqa: E501
                -27*x**3/2 - 27*x**2*y/2 - 27*x**2*z/2 + 18*x**2 + 9*x*y/2 + 9*x*z/2 - 9*x/2,
                27*x*y*z,
                -27*x*y*z - 27*y**2*z - 27*y*z**2 + 27*y*z,
                -27*x**2*z - 27*x*y*z - 27*x*z**2 + 27*x*z,
                -27*x**2*y - 27*x*y**2 - 27*x*y*z + 27*x*y
            ]
        if n == 4:
            return [
                32*x**4/3 + 128*x**3*y/3 + 128*x**3*z/3 - 80*x**3/3 + 64*x**2*y**2 + 128*x**2*y*z - 80*x**2*y + 64*x**2*z**2 - 80*x**2*z + 70*x**2/3 + 128*x*y**3/3 + 128*x*y**2*z - 80*x*y**2 + 128*x*y*z**2 - 160*x*y*z + 140*x*y/3 + 128*x*z**3/3 - 80*x*z**2 + 140*x*z/3 - 25*x/3 + 32*y**4/3 + 128*y**3*z/3 - 80*y**3/3 + 64*y**2*z**2 - 80*y**2*z + 70*y**2/3 + 128*y*z**3/3 - 80*y*z**2 + 140*y*z/3 - 25*y/3 + 32*z**4/3 - 80*z**3/3 + 70*z**2/3 - 25*z/3 + 1,  # noqa: E501
                32*x**4/3 - 16*x**3 + 22*x**2/3 - x,
                32*y**4/3 - 16*y**3 + 22*y**2/3 - y,
                32*z**4/3 - 16*z**3 + 22*z**2/3 - z,
                128*y**3*z/3 - 32*y**2*z + 16*y*z/3,
                64*y**2*z**2 - 16*y**2*z - 16*y*z**2 + 4*y*z,
                128*y*z**3/3 - 32*y*z**2 + 16*y*z/3,
                128*x**3*z/3 - 32*x**2*z + 16*x*z/3,
                64*x**2*z**2 - 16*x**2*z - 16*x*z**2 + 4*x*z,
                128*x*z**3/3 - 32*x*z**2 + 16*x*z/3,
                128*x**3*y/3 - 32*x**2*y + 16*x*y/3,
                64*x**2*y**2 - 16*x**2*y - 16*x*y**2 + 4*x*y,
                128*x*y**3/3 - 32*x*y**2 + 16*x*y/3,
                -128*x**3*z/3 - 128*x**2*y*z - 128*x**2*z**2 + 96*x**2*z - 128*x*y**2*z - 256*x*y*z**2 + 192*x*y*z - 128*x*z**3 + 192*x*z**2 - 208*x*z/3 - 128*y**3*z/3 - 128*y**2*z**2 + 96*y**2*z - 128*y*z**3 + 192*y*z**2 - 208*y*z/3 - 128*z**4/3 + 96*z**3 - 208*z**2/3 + 16*z,  # noqa: E501
                64*x**2*z**2 - 16*x**2*z + 128*x*y*z**2 - 32*x*y*z + 128*x*z**3 - 144*x*z**2 + 28*x*z + 64*y**2*z**2 - 16*y**2*z + 128*y*z**3 - 144*y*z**2 + 28*y*z + 64*z**4 - 128*z**3 + 76*z**2 - 12*z,  # noqa: E501
                -128*x*z**3/3 + 32*x*z**2 - 16*x*z/3 - 128*y*z**3/3 + 32*y*z**2 - 16*y*z/3 - 128*z**4/3 + 224*z**3/3 - 112*z**2/3 + 16*z/3,  # noqa: E501
                -128*x**3*y/3 - 128*x**2*y**2 - 128*x**2*y*z + 96*x**2*y - 128*x*y**3 - 256*x*y**2*z + 192*x*y**2 - 128*x*y*z**2 + 192*x*y*z - 208*x*y/3 - 128*y**4/3 - 128*y**3*z + 96*y**3 - 128*y**2*z**2 + 192*y**2*z - 208*y**2/3 - 128*y*z**3/3 + 96*y*z**2 - 208*y*z/3 + 16*y,  # noqa: E501
                64*x**2*y**2 - 16*x**2*y + 128*x*y**3 + 128*x*y**2*z - 144*x*y**2 - 32*x*y*z + 28*x*y + 64*y**4 + 128*y**3*z - 128*y**3 + 64*y**2*z**2 - 144*y**2*z + 76*y**2 - 16*y*z**2 + 28*y*z - 12*y,  # noqa: E501
                -128*x*y**3/3 + 32*x*y**2 - 16*x*y/3 - 128*y**4/3 - 128*y**3*z/3 + 224*y**3/3 + 32*y**2*z - 112*y**2/3 - 16*y*z/3 + 16*y/3,  # noqa: E501
                -128*x**4/3 - 128*x**3*y - 128*x**3*z + 96*x**3 - 128*x**2*y**2 - 256*x**2*y*z + 192*x**2*y - 128*x**2*z**2 + 192*x**2*z - 208*x**2/3 - 128*x*y**3/3 - 128*x*y**2*z + 96*x*y**2 - 128*x*y*z**2 + 192*x*y*z - 208*x*y/3 - 128*x*z**3/3 + 96*x*z**2 - 208*x*z/3 + 16*x,  # noqa: E501
                64*x**4 + 128*x**3*y + 128*x**3*z - 128*x**3 + 64*x**2*y**2 + 128*x**2*y*z - 144*x**2*y + 64*x**2*z**2 - 144*x**2*z + 76*x**2 - 16*x*y**2 - 32*x*y*z + 28*x*y - 16*x*z**2 + 28*x*z - 12*x,  # noqa: E501
                -128*x**4/3 - 128*x**3*y/3 - 128*x**3*z/3 + 224*x**3/3 + 32*x**2*y + 32*x**2*z - 112*x**2/3 - 16*x*y/3 - 16*x*z/3 + 16*x/3,  # noqa: E501
                128*x**2*y*z - 32*x*y*z,
                128*x*y**2*z - 32*x*y*z,
                128*x*y*z**2 - 32*x*y*z,
                128*x**2*y*z + 256*x*y**2*z + 256*x*y*z**2 - 224*x*y*z + 128*y**3*z + 256*y**2*z**2 - 224*y**2*z + 128*y*z**3 - 224*y*z**2 + 96*y*z,  # noqa: E501
                -128*x*y**2*z + 32*x*y*z - 128*y**3*z - 128*y**2*z**2 + 160*y**2*z + 32*y*z**2 - 32*y*z,
                -128*x*y*z**2 + 32*x*y*z - 128*y**2*z**2 + 32*y**2*z - 128*y*z**3 + 160*y*z**2 - 32*y*z,
                128*x**3*z + 256*x**2*y*z + 256*x**2*z**2 - 224*x**2*z + 128*x*y**2*z + 256*x*y*z**2 - 224*x*y*z + 128*x*z**3 - 224*x*z**2 + 96*x*z,  # noqa: E501
                -128*x**3*z - 128*x**2*y*z - 128*x**2*z**2 + 160*x**2*z + 32*x*y*z + 32*x*z**2 - 32*x*z,
                -128*x**2*z**2 + 32*x**2*z - 128*x*y*z**2 + 32*x*y*z - 128*x*z**3 + 160*x*z**2 - 32*x*z,
                128*x**3*y + 256*x**2*y**2 + 256*x**2*y*z - 224*x**2*y + 128*x*y**3 + 256*x*y**2*z - 224*x*y**2 + 128*x*y*z**2 - 224*x*y*z + 96*x*y,  # noqa: E501
                -128*x**3*y - 128*x**2*y**2 - 128*x**2*y*z + 160*x**2*y + 32*x*y**2 + 32*x*y*z - 32*x*y,
                -128*x**2*y**2 + 32*x**2*y - 128*x*y**3 - 128*x*y**2*z + 160*x*y**2 + 32*x*y*z - 32*x*y,
                -256*x**2*y*z - 256*x*y**2*z - 256*x*y*z**2 + 256*x*y*z
            ]

    raise NotImplementedError


def test_point():
    lagrange = basix.create_element(basix.ElementFamily.P, basix.CellType.point, 0, True)
    assert numpy.allclose(lagrange.tabulate(0, numpy.array([[]])), [[[1]]])
    assert numpy.allclose(lagrange.tabulate(0, numpy.array([[], []])), [[[1, 1]]])


@pytest.mark.parametrize("n", [1, 2, 3, 4, 5])
def test_line(n):
    celltype = basix.CellType.interval
    g = sympy_lagrange(celltype, n)
    x = sympy.Symbol("x")
    lagrange = basix.create_element(basix.ElementFamily.P, celltype, n,
                                    basix.LagrangeVariant.equispaced)
    pts = basix.create_lattice(celltype, 6, basix.LatticeType.equispaced, True)
    nderiv = n
    wtab = lagrange.tabulate(nderiv, pts)
    for k in range(nderiv + 1):
        wsym = numpy.zeros_like(wtab[k])
        for i in range(n + 1):
            wd = sympy.diff(g[i], x, k)
            for j, p in enumerate(pts):
                wsym[j, i] = wd.subs(x, p[0])

        assert numpy.allclose(wtab[k], wsym)


@pytest.mark.parametrize("n", [1, 2])
def test_line_without_variant(n):
    celltype = basix.CellType.interval
    g = sympy_lagrange(celltype, n)
    x = sympy.Symbol("x")
    lagrange = basix.create_element(basix.ElementFamily.P, basix.CellType.interval, n)
    pts = basix.create_lattice(celltype, 6, basix.LatticeType.equispaced, True)
    nderiv = n
    wtab = lagrange.tabulate(nderiv, pts)
    for k in range(nderiv + 1):
        wsym = numpy.zeros_like(wtab[k])
        for i in range(n + 1):
            wd = sympy.diff(g[i], x, k)
            for j, p in enumerate(pts):
                wsym[j, i] = wd.subs(x, p[0])

        assert numpy.allclose(wtab[k], wsym)


@pytest.mark.parametrize("degree", [1, 2, 3, 4, 5])
def test_tri(degree):
    celltype = basix.CellType.triangle
    g = sympy_lagrange(celltype, degree)
    x = sympy.Symbol("x")
    y = sympy.Symbol("y")
    lagrange = basix.create_element(basix.ElementFamily.P, basix.CellType.triangle, degree,
                                    basix.LagrangeVariant.equispaced)
    pts = basix.create_lattice(celltype, 6, basix.LatticeType.equispaced, True)
    nderiv = 3
    wtab = lagrange.tabulate(nderiv, pts)

    for kx in range(nderiv):
        for ky in range(0, nderiv - kx):
            wsym = numpy.zeros_like(wtab[0])
            for i in range(len(g)):
                wd = sympy.diff(g[i], x, kx, y, ky)
                for j, p in enumerate(pts):
                    wsym[j, i] = wd.subs([(x, p[0]), (y, p[1])])

            assert numpy.allclose(wtab[basix.index(kx, ky)], wsym)


@pytest.mark.parametrize("degree", [1, 2, 3, 4])
def test_tet(degree):
    celltype = basix.CellType.tetrahedron
    g = sympy_lagrange(celltype, degree)
    x = sympy.Symbol("x")
    y = sympy.Symbol("y")
    z = sympy.Symbol("z")
    lagrange = basix.create_element(basix.ElementFamily.P, basix.CellType.tetrahedron, degree,
                                    basix.LagrangeVariant.equispaced)
    pts = basix.create_lattice(celltype, 6,
                               basix.LatticeType.equispaced, True)
    nderiv = 1
    wtab = lagrange.tabulate(nderiv, pts)
    for k in range(nderiv + 1):
        for q in range(k + 1):
            for kx in range(q + 1):
                ky = q - kx
                kz = k - q
                wsym = numpy.zeros_like(wtab[0])
                for i in range(len(g)):
                    wd = sympy.diff(g[i], x, kx, y, ky, z, kz)
                    for j, p in enumerate(pts):
                        wsym[j, i] = wd.subs([(x, p[0]),
                                              (y, p[1]),
                                              (z, p[2])])

                assert numpy.allclose(wtab[basix.index(kx, ky, kz)], wsym)


@pytest.mark.parametrize("celltype", [(basix.CellType.interval, basix.CellType.interval),
                                      (basix.CellType.triangle, basix.CellType.triangle),
                                      (basix.CellType.tetrahedron, basix.CellType.tetrahedron)])
@pytest.mark.parametrize("degree", [1, 2, 3, 4])
def test_lagrange(celltype, degree):
    lagrange = basix.create_element(basix.ElementFamily.P, celltype[1], degree,
                                    basix.LagrangeVariant.equispaced)
    pts = basix.create_lattice(celltype[0], 6, basix.LatticeType.equispaced, True)
    w = lagrange.tabulate(0, pts)[0]
    assert(numpy.isclose(numpy.sum(w, axis=1), 1.0).all())


@pytest.mark.parametrize("degree", [1, 2, 3, 4])
def test_dof_transformations_interval(degree):
    lagrange = basix.create_element(basix.ElementFamily.P, basix.CellType.interval, degree,
                                    basix.LagrangeVariant.equispaced)
    assert len(lagrange.base_transformations()) == 0


@pytest.mark.parametrize("degree", [1, 2, 3, 4])
def test_dof_transformations_triangle(degree):
    lagrange = basix.create_element(basix.ElementFamily.P, basix.CellType.triangle, degree,
                                    basix.LagrangeVariant.equispaced)

    permuted = {}
    if degree == 3:
        # Reflect 2 DOFs on edges
        permuted[0] = {3: 4, 4: 3}
        permuted[1] = {5: 6, 6: 5}
        permuted[2] = {7: 8, 8: 7}
    elif degree == 4:
        # Reflect 3 DOFs on edges
        permuted[0] = {3: 5, 5: 3}
        permuted[1] = {6: 8, 8: 6}
        permuted[2] = {9: 11, 11: 9}

    base_transformations = lagrange.base_transformations()
    assert len(base_transformations) == 3

    for i, t in enumerate(base_transformations):
        actual = numpy.zeros_like(t)
        for j, row in enumerate(t):
            if i in permuted and j in permuted[i]:
                actual[j, permuted[i][j]] = 1
            else:
                actual[j, j] = 1
        assert numpy.allclose(t, actual)


@pytest.mark.parametrize("degree", [1, 2, 3, 4])
def test_dof_transformations_tetrahedron(degree):
    lagrange = basix.create_element(basix.ElementFamily.P, basix.CellType.tetrahedron, degree,
                                    basix.LagrangeVariant.equispaced)

    permuted = {}
    if degree == 3:
        # Reflect 2 DOFs on edges
        permuted[0] = {4: 5, 5: 4}
        permuted[1] = {6: 7, 7: 6}
        permuted[2] = {8: 9, 9: 8}
        permuted[3] = {10: 11, 11: 10}
        permuted[4] = {12: 13, 13: 12}
        permuted[5] = {14: 15, 15: 14}
    elif degree == 4:
        # Reflect 3 DOFs on edges
        permuted[0] = {4: 6, 6: 4}
        permuted[1] = {7: 9, 9: 7}
        permuted[2] = {10: 12, 12: 10}
        permuted[3] = {13: 15, 15: 13}
        permuted[4] = {16: 18, 18: 16}
        permuted[5] = {19: 21, 21: 19}
        # Rotate and reflect 3 DOFs on faces
        permuted[6] = {22: 24, 23: 22, 24: 23}
        permuted[7] = {23: 24, 24: 23}
        permuted[8] = {25: 27, 26: 25, 27: 26}
        permuted[9] = {26: 27, 27: 26}
        permuted[10] = {28: 30, 29: 28, 30: 29}
        permuted[11] = {29: 30, 30: 29}
        permuted[12] = {31: 33, 32: 31, 33: 32}
        permuted[13] = {32: 33, 33: 32}

    base_transformations = lagrange.base_transformations()
    assert len(base_transformations) == 14

    for i, t in enumerate(base_transformations):
        actual = numpy.zeros_like(t)
        for j, row in enumerate(t):
            if i in permuted and j in permuted[i]:
                actual[j, permuted[i][j]] = 1
            else:
                actual[j, j] = 1
        assert numpy.allclose(t, actual)


@pytest.mark.parametrize("degree", [1, 2, 3, 4])
@pytest.mark.parametrize("celltype", [
    basix.CellType.quadrilateral,
    basix.CellType.hexahedron,
    basix.CellType.pyramid,
    basix.CellType.prism
])
def test_celltypes(degree, celltype):
    tp = basix.create_element(basix.ElementFamily.P, celltype, degree,
                              basix.LagrangeVariant.equispaced)
    pts = basix.create_lattice(celltype, 5,
                               basix.LatticeType.equispaced, True)
    w = tp.tabulate(0, pts)[0]
    assert(numpy.allclose(numpy.sum(w, axis=1), 1.0))


def leq(a, b):
    return a <= b or numpy.isclose(a, b)


def in_cell(celltype, p):
    if celltype == basix.CellType.interval:
        return leq(0, p[0]) and leq(p[0], 1)
    if celltype == basix.CellType.triangle:
        return leq(0, p[0]) and leq(0, p[1]) and leq(p[0] + p[1], 1)
    if celltype == basix.CellType.tetrahedron:
        return leq(0, p[0]) and leq(0, p[1]) and leq(0, p[2]) and leq(p[0] + p[1] + p[2], 1)
    if celltype == basix.CellType.quadrilateral:
        return leq(0, p[0]) and leq(0, p[1]) and leq(p[0], 1) and leq(p[1], 1)
    if celltype == basix.CellType.hexahedron:
        return leq(0, p[0]) and leq(0, p[1]) and leq(0, p[2]) and leq(p[0], 1) and leq(p[1], 1) and leq(p[2], 1)
    if celltype == basix.CellType.prism:
        return leq(0, p[0]) and leq(0, p[1]) and leq(0, p[2]) and leq(p[0] + p[1], 1) and leq(p[2], 1)


@pytest.mark.parametrize("variant", [
    basix.LagrangeVariant.equispaced,
    basix.LagrangeVariant.gll_warped, basix.LagrangeVariant.gll_isaac, basix.LagrangeVariant.gll_centroid,
    basix.LagrangeVariant.chebyshev_warped, basix.LagrangeVariant.chebyshev_isaac,
    basix.LagrangeVariant.chebyshev_centroid,
    basix.LagrangeVariant.gl_warped, basix.LagrangeVariant.gl_isaac, basix.LagrangeVariant.gl_centroid,
    basix.LagrangeVariant.vtk
])
@pytest.mark.parametrize("celltype", [
    basix.CellType.triangle, basix.CellType.tetrahedron,
    basix.CellType.quadrilateral, basix.CellType.hexahedron,
])
@pytest.mark.parametrize("degree", range(1, 5))
def test_variant_points(celltype, degree, variant):
    e = basix.create_element(basix.ElementFamily.P, celltype, degree, variant, True)

    for p in e.points:
        assert in_cell(celltype, p)


@pytest.mark.parametrize("variant", [
    basix.LagrangeVariant.chebyshev_warped, basix.LagrangeVariant.chebyshev_isaac,
    basix.LagrangeVariant.chebyshev_centroid,
    basix.LagrangeVariant.gl_warped, basix.LagrangeVariant.gl_isaac, basix.LagrangeVariant.gl_centroid,
])
@pytest.mark.parametrize("celltype", [
    basix.CellType.triangle, basix.CellType.tetrahedron,
    basix.CellType.quadrilateral, basix.CellType.hexahedron,
])
def test_continuous_lagrange(celltype, variant):
    # The variants used in this test can only be used for discontinuous Lagrange,
    # so trying to create them should throw a runtime error
    with pytest.raises(RuntimeError):
        basix.create_element(basix.ElementFamily.P, celltype, 4, variant, False)


@pytest.mark.parametrize("celltype", [
    basix.CellType.interval, basix.CellType.triangle, basix.CellType.tetrahedron,
    basix.CellType.quadrilateral, basix.CellType.hexahedron,
])
@pytest.mark.parametrize("degree", range(1, 9))
def test_vtk_element(celltype, degree):
    if degree > 5 and celltype == basix.CellType.hexahedron:
        pytest.skip("Skipping slow test on hexahedron")

    equi = basix.create_element(basix.ElementFamily.P, celltype, degree, basix.LagrangeVariant.equispaced, True)
    vtk = basix.create_element(basix.ElementFamily.P, celltype, degree, basix.LagrangeVariant.vtk, True)

    assert vtk.points.shape == equi.points.shape

    perm = []

    for i, p in enumerate(vtk.points):
        for j, q in enumerate(vtk.points):
            if i != j:
                assert not numpy.allclose(p, q)

        for j, q in enumerate(equi.points):
            if numpy.allclose(p, q):
                perm.append(j)
                break
        else:
            raise ValueError(f"Incorrect point in VTK variant: {p}")

    # Test against permutations that were previously in DOLFINx
    if celltype == basix.CellType.triangle:
        if degree <= 9:
            target = [0, 1, 2]
            j = 3
            target += [2 * degree + k for k in range(1, degree)]
            target += [2 + k for k in range(1, degree)]
            target += [2 * degree + 1 - k for k in range(1, degree)]

            if degree == 3:
                target += [len(target) + i for i in [0]]
            elif degree == 4:
                target += [len(target) + i for i in [0, 1, 2]]
            elif degree == 5:
                target += [len(target) + i for i in [0, 2, 5, 1, 4, 3]]
            elif degree == 6:
                target += [len(target) + i for i in [0, 3, 9, 1, 2, 6, 8, 7, 4, 5]]
            elif degree == 7:
                target += [len(target) + i for i in [0, 4, 14, 1, 2, 3, 8, 11, 13, 12, 9, 5, 6, 7, 10]]
            elif degree == 8:
                target += [len(target) + i for i in [0, 5, 20, 1, 2, 3, 4, 10, 14, 17, 19,
                                                     18, 15, 11, 6, 7, 9, 16, 8, 13, 12]]
            elif degree == 9:
                target += [len(target) + i for i in [0, 6, 27, 1, 2, 3, 4, 5, 12, 17, 21, 24, 26, 25,
                                                     22, 18, 13, 7, 8, 11, 23, 9, 10, 16, 20, 19, 14, 15]]

            assert perm == target

    elif celltype == basix.CellType.tetrahedron:
        if degree == 1:
            assert perm == [0, 1, 2, 3]
        elif degree == 2:
            assert perm == [0, 1, 2, 3, 9, 6, 8, 7, 5, 4]
        elif degree == 3:
            assert perm == [0, 1, 2, 3, 14, 15, 8, 9, 13, 12,
                            10, 11, 6, 7, 4, 5, 18, 16, 17, 19]

    elif celltype == basix.CellType.quadrilateral:
        target = [0, 1, 3, 2]
        target += [4 + k for k in range(degree - 1)]
        target += [4 + 2 * (degree - 1) + k for k in range(degree - 1)]
        target += [4 + 3 * (degree - 1) + k for k in range(degree - 1)]
        target += [4 + (degree - 1) + k for k in range(degree - 1)]
        target += [4 + (degree - 1) * 4 + k for k in range((degree - 1) ** 2)]

        assert target == perm

    elif celltype == basix.CellType.hexahedron:
        if degree == 1:
            assert perm == [0, 1, 3, 2, 4, 5, 7, 6]
        elif degree == 2:
            assert perm == [0, 1, 3, 2, 4, 5, 7, 6, 8, 11, 13, 9, 16, 18,
                            19, 17, 10, 12, 15, 14, 22, 23, 21, 24, 20, 25, 26]


@pytest.mark.parametrize("variant", [
    basix.LagrangeVariant.legendre,
])
@pytest.mark.parametrize("celltype", [
    basix.CellType.interval,
    basix.CellType.triangle, basix.CellType.tetrahedron,
    basix.CellType.quadrilateral, basix.CellType.hexahedron,
])
@pytest.mark.parametrize("degree", range(1, 5))
def test_legendre(celltype, degree, variant):
    e = basix.create_element(basix.ElementFamily.P, celltype, degree, variant, True)
    for p in e.points:
        assert in_cell(celltype, p)


@pytest.mark.parametrize("variant", [
    basix.DPCVariant.simplex_equispaced,
    basix.DPCVariant.simplex_gll,
    basix.DPCVariant.horizontal_equispaced,
    basix.DPCVariant.horizontal_gll,
    basix.DPCVariant.diagonal_equispaced,
    basix.DPCVariant.diagonal_gll,
])
@pytest.mark.parametrize("celltype", [
    basix.CellType.quadrilateral, basix.CellType.hexahedron,
])
@pytest.mark.parametrize("degree", range(5))
def test_dpc(celltype, degree, variant):
    e = basix.create_element(basix.ElementFamily.DPC, celltype, degree, variant, True)
    for p in e.points:
        assert in_cell(celltype, p)


@pytest.mark.parametrize("celltype", [
    basix.CellType.interval,
    basix.CellType.triangle, basix.CellType.tetrahedron,
    basix.CellType.quadrilateral, basix.CellType.hexahedron,
])
@pytest.mark.parametrize("degree", range(1, 5))
def test_legendre_lagrange_variant(celltype, degree):
    e = basix.create_element(
        basix.ElementFamily.P, celltype, degree, basix.LagrangeVariant.legendre, True)

    # Test that the basis functions are orthogonal
    pts, wts = basix.make_quadrature(celltype, degree * 2)
    values = e.tabulate(0, pts)[0, :, :, 0].T
    for i, row_i in enumerate(values):
        for j, row_j in enumerate(values):
            integral = numpy.sum(row_i * row_j * wts)
            if i == j:
                assert numpy.isclose(integral, 1)
            else:
                assert numpy.isclose(integral, 0)

    # Test that the basis function span the correct set
    pts = basix.create_lattice(celltype, 2 * (degree + 1), basix.LatticeType.equispaced, True)
    values = e.tabulate(0, pts)[0, :, :, 0]
    i_pts = e.points
    i_mat = e.interpolation_matrix

    tdim = len(basix.topology(celltype)) - 1
    if tdim == 1:
        for px in range(degree + 1):
            evals = i_pts[:, 0] ** px
            coeffs = i_mat @ evals
            computed_values = [numpy.dot(v, coeffs) for v in values]
            actual_values = pts[:, 0] ** px
            assert numpy.allclose(computed_values, actual_values)

    elif tdim == 2:
        powers = []
        if celltype == basix.CellType.triangle:
            for px in range(degree + 1):
                for py in range(degree + 1 - px):
                    powers.append((px, py))
        if celltype == basix.CellType.quadrilateral:
            for px in range(degree + 1):
                for py in range(degree + 1):
                    powers.append((px, py))
        for px, py in powers:
            evals = i_pts[:, 0] ** px * i_pts[:, 1] ** py
            coeffs = i_mat @ evals
            computed_values = [numpy.dot(v, coeffs) for v in values]
            actual_values = pts[:, 0] ** px * pts[:, 1] ** py
            assert numpy.allclose(computed_values, actual_values)

    else:
        assert tdim == 3
        powers = []
        if celltype == basix.CellType.tetrahedron:
            for px in range(degree + 1):
                for py in range(degree + 1 - px):
                    for pz in range(degree + 1 - px - py):
                        powers.append((px, py, pz))
        if celltype == basix.CellType.hexahedron:
            for px in range(degree + 1):
                for py in range(degree + 1):
                    for pz in range(degree + 1):
                        powers.append((px, py, pz))
        for px, py, pz in powers:
            evals = i_pts[:, 0] ** px * i_pts[:, 1] ** py * i_pts[:, 2] ** pz
            coeffs = i_mat @ evals
            computed_values = [numpy.dot(v, coeffs) for v in values]
            actual_values = pts[:, 0] ** px * pts[:, 1] ** py * pts[:, 2] ** pz
            assert numpy.allclose(computed_values, actual_values)


@pytest.mark.parametrize("celltype", [
    basix.CellType.quadrilateral, basix.CellType.hexahedron,
])
@pytest.mark.parametrize("degree", range(1, 5))
def test_legendre_dpc_variant(celltype, degree):
    e = basix.create_element(
        basix.ElementFamily.DPC, celltype, degree, basix.DPCVariant.legendre, True)

    # Test that the basis functions are orthogonal
    pts, wts = basix.make_quadrature(celltype, degree * 2)
    values = e.tabulate(0, pts)[0, :, :, 0].T
    for i, row_i in enumerate(values):
        for j, row_j in enumerate(values):
            integral = numpy.sum(row_i * row_j * wts)
            if i == j:
                assert numpy.isclose(integral, 1)
            else:
                assert numpy.isclose(integral, 0)

    # Test that the basis function span the correct set
    pts = basix.create_lattice(celltype, 2 * (degree + 1), basix.LatticeType.equispaced, True)
    values = e.tabulate(0, pts)[0, :, :, 0]
    i_pts = e.points
    i_mat = e.interpolation_matrix

    tdim = len(basix.topology(celltype)) - 1
    if tdim == 2:
        for px in range(degree + 1):
            for py in range(degree + 1 - px):
                evals = i_pts[:, 0] ** px * i_pts[:, 1] ** py
                coeffs = i_mat @ evals
                computed_values = [numpy.dot(v, coeffs) for v in values]
                actual_values = pts[:, 0] ** px * pts[:, 1] ** py
                assert numpy.allclose(computed_values, actual_values)

    else:
        assert tdim == 3
        for px in range(degree + 1):
            for py in range(degree + 1 - px):
                for pz in range(degree + 1 - px - py):
                    evals = i_pts[:, 0] ** px * i_pts[:, 1] ** py * i_pts[:, 2] ** pz
                    coeffs = i_mat @ evals
                    computed_values = [numpy.dot(v, coeffs) for v in values]
                    actual_values = pts[:, 0] ** px * pts[:, 1] ** py * pts[:, 2] ** pz
                    assert numpy.allclose(computed_values, actual_values)



import os
import pytest

# Get all the demos in this folder
path = os.path.dirname(os.path.realpath(__file__))
demos = []
for folder in os.listdir(path):
    if folder.startswith("demo_"):
        subpath = os.path.join(path, folder)
        if os.path.isdir(subpath) and os.path.isfile(os.path.join(subpath, "main.cpp")):
            demos.append(folder)


@pytest.mark.parametrize("demo", demos)
def test_demo(demo):
    demo_build = f"{path}/{demo}/_build"
    assert os.system(f"mkdir -p {demo_build} && cd {demo_build} && cmake .. && make && ./{demo}") == 0



"""Generate docs for the pybind layer from the C++ header files."""

import os
import re

path = os.path.dirname(os.path.realpath(__file__))
cpp_path = os.path.join(path, "../cpp/basix")

replacements = [(";", "@semicolon@"), ("{", "@openbrace@"), ("}", "@closebrace@"),
                ("<", "@opentri@"), (">", "@closetri@"), ("(", "@openb@"), (")", "@closeb@"),
                ("[", "@opensq@"), ("]", "@closesq@")]


def replace(txt):
    for i, j in replacements:
        txt = txt.replace(i, j)
    return txt


def unreplace(txt):
    for i, j in replacements:
        txt = txt.replace(j, i)
    return txt


def remove_types(matches):
    """Remove the types from a function declaration."""
    vars = [i.strip().split(" ")[-1].split("=")[0] for i in matches[1].split(",")]
    return "(" + ", ".join(vars) + ")"


def prepare_cpp(content):
    """Prepare a cpp file for parsing."""
    out = ""

    for line in content.split("\n"):
        if not line.strip().startswith("#"):
            if line.strip().startswith("//"):
                out += replace(line) + "\n"
            else:
                out += line + "\n"

    out = re.sub(r"namespace [^{]*\{", ";", out)
    out = re.sub(r"class [^{]*\{", "", out)

    while "{" in out:
        out = re.sub(r"\{[^{}]*\}", ";", out)
    while "<" in out and ">" in out:
        out = re.sub(r"<[^<>]*>", "", out)
    out = re.sub(r"\(([^\)]*)\)", remove_types, out)
    return out


def get_docstring(matches):
    """Get documentation from a header file."""
    # Parse the info
    info = matches[1]
    typename = None
    if " : " in info:
        info, typename = info.split(" : ")
    file, function, info = info.split(" > ", 2)
    if " > " in info:
        info_type, info = info.split(" > ")
    else:
        info_type = info

    # Read documentation from a header file
    with open(os.path.join(cpp_path, file)) as f:
        content = prepare_cpp(f.read())

    if "::" in function:
        function = function.split("::")[-1]
    if "(" not in function:
        function += "("

    if function not in content:
        print(function)
    assert function in content
    doc = content.split(function)[0].split(";")[-1]

    # Convert doxygen syntax to Python docs

    doc = "\n".join([i.strip()[4:] for i in doc.split("\n") if i.strip().startswith("///")])
    doc = doc.replace("@f$", "$")
    doc = doc.replace("@f[", "\\[")
    doc = doc.replace("@f]", "\\]")
    doc = doc.replace("@note", "NOTE:")
    doc = doc.replace("@todo", "TODO:")
    doc = doc.replace("@warning", "WARNING:")
    doc = unreplace(doc)

    if info_type == "doc":
        assert typename is None
        docstring = doc.split("@param")[0].split("@return")[0].strip()
        doclines = docstring.split("\n")
        in_code = False
        for i, j in enumerate(doclines):
            if re.match(r"^~+$", j):
                if in_code:
                    in_code = False
                    doclines[i] = "\n"
                else:
                    in_code = True
                    doclines[i] = "\n.. code-block::\n"
            elif in_code:
                doclines[i] = " " + doclines[i]
            else:
                doclines[i] = doclines[i].strip()
        for i, j in enumerate(doclines[:-1]):
            if re.match(r"^-+$", doclines[i + 1]):
                doclines[i] = f"*{j}*"
                doclines[i + 1] = ""
        return "\n".join(doclines)

    if info_type == "param":
        assert typename is not None
        params = {}
        for i in doc.split("@param")[1:]:
            i = i.split("@return")[0]
            i = i.replace("[in]", "")
            i = i.replace("[out]", "")
            i = i.replace("[in,out]", "")
            i = " ".join([j.strip() for j in i.strip().split("\n")])
            if " " in i:
                p, pdoc = i.split(" ", 1)
            else:
                p = i
                pdoc = "TODO: document this"
            params[p] = "\n        ".join(pdoc.split("\n"))
        return f"{info} ({typename}): {params[info]}"

    if info_type == "return":
        assert typename is not None
        returns = [i.split("@param")[0].strip() for i in doc.split("@return")[1:]]
        if len(returns) == 0:
            returns.append("TODO: document this")
        assert len(returns) == 1
        returns = "\n    ".join(returns[0].split("\n"))
        return f"{typename}: {returns}"


def generate_docs():
    with open(os.path.join(path, "docs.h.template")) as f:
        docs = f.read()

    docs = docs.replace("{{DOCTYPE}}", "const std::string")
    docs = re.sub(r"\{\{(.+?)\}\}", get_docstring, docs)

    return docs


if __name__ == "__main__":
    with open(os.path.join(path, "docs.h"), "w") as f:
        f.write(generate_docs())



"""Functions for creating finite elements."""

from ._basixcpp import ElementFamily as _EF
from ._basixcpp import FiniteElement  # noqa: F401


def string_to_family(family: str, cell: str) -> _EF:
    """Get a Basix ElementFamily enum representing the family type on the given cell.

    Args:
        family: The element family as a string.
        cell: The cell type as a string.

    Returns:
        The element family.
    """
    # Family names that are valid for all cells
    families = {
        "Lagrange": _EF.P,
        "P": _EF.P,
        "Bubble": _EF.bubble,
        "bubble": _EF.bubble,
    }
    # Family names that are valid on non-interval cells
    if cell != "interval":
        families.update({
            "RT": _EF.RT,
            "Raviart-Thomas": _EF.RT,
            "N1F": _EF.RT,
            "N1div": _EF.RT,
            "Nedelec 1st kind H(div)": _EF.RT,
            "N1E": _EF.N1E,
            "N1curl": _EF.N1E,
            "Nedelec 1st kind H(curl)": _EF.N1E,
            "BDM": _EF.BDM,
            "Brezzi-Douglas-Marini": _EF.BDM,
            "N2F": _EF.BDM,
            "N2div": _EF.BDM,
            "Nedelec 2nd kind H(div)": _EF.BDM,
            "N2E": _EF.N2E,
            "N2curl": _EF.N2E,
            "Nedelec 2nd kind H(curl)": _EF.N2E,
        })
    # Family names that are valid for intervals
    if cell == "interval":
        families.update({
            "DPC": _EF.P,
        })
    # Family names that are valid for tensor product cells
    if cell in ["interval", "quadrilateral", "hexahedron"]:
        families.update({
            "Q": _EF.P,
            "Serendipity": _EF.serendipity,
            "serendipity": _EF.serendipity,
            "S": _EF.serendipity,
        })
    # Family names that are valid for quads and hexes
    if cell in ["quadrilateral", "hexahedron"]:
        families.update({
            "RTCF": _EF.RT,
            "DPC": _EF.DPC,
            "NCF": _EF.RT,
            "RTCE": _EF.N1E,
            "NCE": _EF.N1E,
            "BDMCF": _EF.BDM,
            "BDMCE": _EF.N2E,
            "AAF": _EF.BDM,
            "AAE": _EF.N2E,
        })
    # Family names that are valid for triangles and tetrahedra
    if cell in ["triangle", "tetrahedron"]:
        families.update({
            "Regge": _EF.Regge,
            "CR": _EF.CR,
            "Crouzeix-Raviart": _EF.CR,
        })
    # Family names that are valid for triangles
    if cell in "triangle":
        families.update({
            "HHJ": _EF.HHJ,
            "Hellan-Herrmann-Johnson": _EF.HHJ,
        })

    if family in families:
        return families[family]

    raise ValueError(f"Unknown element family: {family} with cell type {cell}")



import argparse
from markdown import markdown as _markdown
import os
import basix
import re
import yaml

_path = os.path.dirname(os.path.realpath(__file__))


def path(dir):
    return os.path.join(_path, dir)


temp = path("_temp")

parser = argparse.ArgumentParser(description="Build Basix documentation")
parser.add_argument('--url', metavar='url',
                    default="http://localhost", help="URL of built documentation")
parser.add_argument('--clone', metavar='clone',
                    default="true", help="Clone the web repository")

args = parser.parse_args()
url = args.url
url_no_http = url.split("//")[1]


def markdown(txt):
    """Convert markdown to HTML."""
    # Convert code inside ```s (as these are not handled by the markdown library
    tsp = txt.split("```")
    out = ""
    code = False
    for part in tsp:
        if code:
            part = part.strip().split("\n")
            if part[0] in ["console", "bash", "python", "c", "cpp"]:
                part = part[1:]
            part = "\n".join(part)
            out += f"<pre><code>{part}</code></pre>"
        else:
            out += part
        code = not code
    return _markdown(out)


def jekyll(a):
    """
    Convert Jekyll to HTML.

    This is used to convert template pages from the FEniCS website to the format needed
    here.
    """
    while "{% include" in a:
        b, c = a.split("{% include", 1)
        c, d = c.split("%}", 1)
        with open(path(f"web/_includes/{c.strip()}")) as f:
            a = b + f.read() + d

    while "{% for link in site.data.navbar %}" in a:
        b, c = a.split("{% for link in site.data.navbar %}", 1)
        c, d = c.split("{% endfor %}", 1)
        with open(path("web/_data/navbar.yml")) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        a = b
        for i in data:
            this_c = c
            for j, k in i.items():
                if j == "page" and k.startswith("/"):
                    print(j, k)
                    this_c = this_c.replace(f"{{{{ link.{j} }}}}", f"https://fenicsproject.org{k}")
                else:
                    this_c = this_c.replace(f"{{{{ link.{j} }}}}", k)
            a += this_c
        a += d
    if "{% if page.image %}" in a:
        b, c = a.split("{% if page.image %}")
        _, c = c.split("{% else %}", 1)
        c, d = c.split("{% endif %}", 1)
        a = b + c + d

    if "{{ page.title | default: site.title }}" in a:
        with open(path("template/title.html")) as f:
            a = a.replace("{{ page.title | default: site.title }}", f.read())
    if "\n{% if page.subtitle %}" in a:
        b, c = a.split("\n{% if page.subtitle %}")
        _, c = c.split("\n{% endif %}", 1)
        a = b + c

    a = re.sub(r"\{\{[^}]+\| default: (?:'|\")([^}]+)(?:'|\") \}\}", r"\1", a)
    a = re.sub(r"\{\{ (?:'|\")\/assets\/css\/style\.css\?v=(?:'|\")[^\}]+\}\}", "/assets/css/style.css", a)
    a = a.replace("{% seo %}", "{{SEO}}")
    a = re.sub(r"([\"'\(])/assets", r"\1https://fenicsproject.org/assets", a)

    assert "{%" not in a

    return a


def unicode_to_html(txt):
    txt = txt.replace(u"\xe9", "&eacute;")
    return txt


def make_html_page(html):
    with open(os.path.join(temp, "intro.html")) as f:
        pre = f.read()
    with open(os.path.join(temp, "outro.html")) as f:
        post = f.read()
    return pre + html + post


def link_markup(matches):
    return f"[{url_no_http}/{matches[1]}]({url}/{matches[1]})"


def insert_info(txt):
    if "{{SUMMARY}}" in txt:
        started = False
        info = ""
        with open(path("../../README.md")) as f:
            for line in f:
                if not started and line.strip() != "" and line[0] not in ["!", "#"]:
                    started = True
                if started:
                    if line.startswith("#"):
                        break
                    info += line
        info = info.replace("joss/img", "img")
        txt = txt.replace("{{SUMMARY}}", info)
    if "{{SUPPORTED ELEMENTS}}" in txt:
        with open(path("../../README.md")) as f:
            info = "## Supported elements" + f.read().split("## Supported elements")[1].split("\n## ")[0]
        info = info.replace("joss/img", "img")
        txt = txt.replace("{{SUPPORTED ELEMENTS}}", info)
    if "{{INSTALL}}" in txt:
        with open(path("../../INSTALL.md")) as f:
            info = f.read()
        txt = txt.replace("{{INSTALL}}", info)
    txt = txt.replace("{{URL}}", url)
    txt = txt.replace("{{VERSION}}", basix.__version__)
    txt = re.sub(r"\{\{link:URL/([^\}]*)\}\}", link_markup, txt)
    return txt


def system(command):
    assert os.system(command) == 0


# Make paths
for dir in ["html", "cpp", "python"]:
    if os.path.isdir(path(dir)):
        system(f"rm -r {path(dir)}")
if os.path.isdir(temp):
    system(f"rm -r {temp}")
system(f"mkdir {path('html')}")
system(f"mkdir {temp}")

# Copy cpp and python folders
system(f"cp -r {path('../cpp')} {path('cpp')}")
system(f"cp -r {path('../python')} {path('python')}")
system(f"mkdir {path('python/source/_templates')}")

# Prepare templates
if args.clone == "true":
    if os.path.isdir(path('web')):
        system(f"rm -r {path('web')}")
    system(f"git clone http://github.com/FEniCS/web {path('web')}")

with open(path('web/_layouts/default.html')) as f:
    intro, outro = f.read().split("\n    {{ title }}\n    {{ content }}\n")
    intro = insert_info(jekyll(intro))
    outro = insert_info(jekyll(outro))

with open(path("template/navbar.html")) as f:
    intro += f"<h2 id=\"project_subtitle\">{insert_info(f.read())}</h2>"

with open(os.path.join(temp, "intro.html"), "w") as f:
    f.write(intro.replace("{{SEO}}", "<title>Basix documentation</title>"))
with open(os.path.join(temp, "outro.html"), "w") as f:
    f.write(outro)

intro += "\n".join(outro.split("\n")[:2])
outro = "\n".join(outro.split("\n")[2:])

with open(path("cpp/footer.html"), "w") as f:
    f.write(outro)
with open(path("cpp/header.html"), "w") as f:
    with open(path("cpp-seo.html")) as f2:
        f.write(insert_info(intro.replace("{{SEO}}", f2.read())))
    with open(path("cpp-header.html")) as f2:
        f.write(insert_info(f2.read()))

content = "{% extends \"!layout.html\" %}\n"
content += "{%- block content %}\n"
content += intro.split("<body>")[1] + "\n"
content += "    {{ super() }}\n"
content += outro.split("</body>")[0] + "\n"
content += "{% endblock %}\n"
content += "{%- block extrahead %}\n"
content += intro.split("<head>")[1].split("</head>")[0] + "\n"
content += "  {{ super() }}\n"
content += "{% endblock %}"

with open(path("python/source/_templates/layout.html"), "w") as f:
    f.write(insert_info(content))

with open(path("cpp/Doxyfile")) as f:
    content = ""
    for line in f:
        if line.startswith("HTML_HEADER"):
            content += f"{line.split('=')[0]}= header.html\n"
        elif line.startswith("HTML_FOOTER"):
            content += f"{line.split('=')[0]}= footer.html\n"
        else:
            content += line.replace(" ..", " ../..")

with open(path("cpp/Doxyfile"), "w") as f:
    f.write(content)

# Copy images and assets
system(f"cp -r {path('../../joss/img')} {path('html')}/img")
system(f"cp -r {path('assets')} {path('html')}/assets")

# Convert markdown to html
for file in os.listdir(_path):
    if file.endswith(".md"):
        with open(path(file)) as f:
            contents = unicode_to_html(insert_info(f.read()))
        with open(os.path.join(path('html'), file[:-3] + ".html"), "w") as f:
            f.write(make_html_page(markdown(contents)))

# Make cpp docs
system(f"cd {path('cpp')} && doxygen")
system(f"cp -r {path('cpp/html')} {path('html/cpp')}")

# Make demos
os.system(f"rm {path('../../demo/python/*.py.rst')}")
os.system(f"rm {path('../../demo/python/*.png')}")
# If file saves matplotlib images, run the demo
for file in os.listdir(path("../../demo/python")):
    if file.endswith(".py"):
        with open(path(f"../../demo/python/{file}")) as f:
            content = f.read()
        if "savefig" in content:
            here = os.getcwd()
            os.chdir(path("../../demo/python"))
            system(f"python3 {file}")
            os.chdir(here)
system(f"cd {path('../../demo/python')} && python3 convert_to_rst.py")
system(f"mkdir {path('python/source/demo')}")
system(f"cp {path('../../demo/python/*.rst')} {path('python/source/demo')}")
system(f"cp {path('../../demo/python/*.png')} {path('python/source/demo')}")
for file in os.listdir(path("python/source/demo")):
    if file.endswith(".py.rst"):
        with open(path(f"python/source/demo/{file}")) as f:
            content = f.read()
        content = content.replace("literalinclude:: ../", "literalinclude:: ../../../../../demo/")
        with open(path(f"python/source/demo/{file}"), "w") as f:
            f.write(content)

with open(path("python/source/index.rst"), "a") as f:
    f.write("\n")
    f.write(".. toctree::\n")
    f.write("   demo/index")

# Make python docs
system(f"cd {path('python')} && make html")
system(f"cp -r {path('python/build/html')} {path('html/python')}")



import pytest
import basix
import basix.ufl_wrapper


@pytest.mark.parametrize("inputs", [
    ("Lagrange", "triangle", 2),
    ("Lagrange", basix.CellType.triangle, 2),
    (basix.ElementFamily.P, basix.CellType.triangle, 2),
    (basix.ElementFamily.P, "triangle", 2),
])
def test_create_element(inputs):
    basix.ufl_wrapper.create_element(*inputs)


@pytest.mark.parametrize("inputs", [
    ("Lagrange", "triangle", 2),
    ("Lagrange", basix.CellType.triangle, 2),
    (basix.ElementFamily.P, basix.CellType.triangle, 2),
    (basix.ElementFamily.P, "triangle", 2),
])
def test_create_vector_element(inputs):
    basix.ufl_wrapper.create_vector_element(*inputs)


@pytest.mark.parametrize("inputs", [
    ("Lagrange", "triangle", 2),
    ("Lagrange", basix.CellType.triangle, 2),
    (basix.ElementFamily.P, basix.CellType.triangle, 2),
    (basix.ElementFamily.P, "triangle", 2),
])
def test_create_tensor_element(inputs):
    basix.ufl_wrapper.create_tensor_element(*inputs)
